__all__ = ["PluginManager", "get_plugin_manager"]

import asyncio
from collections import defaultdict
from collections.abc import Awaitable, Callable, Coroutine
import contextlib
import copy
from importlib.metadata import EntryPoint, EntryPoints, entry_points
from inspect import iscoroutinefunction
import logging
import sys
import traceback
from typing import Any, Generic, NamedTuple, Optional, TypeAlias, TypeVar, cast
import warnings

from dacite import Config as DaciteConfig, DaciteError, from_dict

from amqtt.contexts import Action, BaseContext
from amqtt.errors import PluginCoroError, PluginImportError, PluginInitError, PluginLoadError
from amqtt.events import BrokerEvents, Events, MQTTEvents
from amqtt.plugins.base import BaseAuthPlugin, BasePlugin, BaseTopicPlugin
from amqtt.session import Session
from amqtt.utils import import_string


class Plugin(NamedTuple):
    name: str
    ep: EntryPoint
    object: Any


plugins_manager: dict[str, "PluginManager[Any]"] = {}


def get_plugin_manager(namespace: str) -> "PluginManager[Any] | None":
    """Get the plugin manager for a given namespace.

    :param namespace: The namespace of the plugin manager to retrieve.
    :return: The plugin manager for the given namespace, or None if it doesn't exist.
    """
    return plugins_manager.get(namespace)


def safe_issubclass(sub_class: Any, super_class: Any) -> bool:
    try:
        return issubclass(sub_class, super_class)
    except TypeError:
        return False


AsyncFunc: TypeAlias = Callable[..., Coroutine[Any, Any, None]]
C = TypeVar("C", bound=BaseContext)


class PluginManager(Generic[C]):
    """Wraps contextlib Entry point mechanism to provide a basic plugin system.

    Plugins are loaded for a given namespace (group). This plugin manager uses coroutines to
    run plugin calls asynchronously in an event queue.
    """

    def __init__(self, namespace: str, context: C | None, loop: asyncio.AbstractEventLoop | None = None) -> None:
        try:
            self._loop = loop if loop is not None else asyncio.get_running_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self.logger = logging.getLogger(namespace)
        self.context = context if context is not None else BaseContext()
        self.context.loop = self._loop
        self._plugins: list[BasePlugin[C]] = []
        self._auth_plugins: list[BaseAuthPlugin] = []
        self._topic_plugins: list[BaseTopicPlugin] = []
        self._event_plugin_callbacks: dict[str, list[AsyncFunc]] = defaultdict(list)
        self._is_topic_filtering_enabled = False
        self._is_auth_filtering_enabled = False

        self._load_plugins(namespace)
        self._fired_events: list[asyncio.Future[Any]] = []
        plugins_manager[namespace] = self

    @property
    def app_context(self) -> BaseContext:
        return self.context

    def _load_plugins(self, namespace: str | None = None) -> None:
        """Load plugins from entrypoint or config dictionary.

        config style is now recommended; entrypoint has been deprecated
        Example:
            config = {
                'listeners':...,
                'plugins': {
                    'myproject.myfile.MyPlugin': {}
            }
        """
        if self.app_context.config and self.app_context.config.get("plugins", None) is not None:
            # plugins loaded directly from config dictionary

            if "auth" in self.app_context.config and self.app_context.config["auth"] is not None:
                self.logger.warning("Loading plugins from config will ignore 'auth' section of config")
            if "topic-check" in self.app_context.config and self.app_context.config["topic-check"] is not None:
                self.logger.warning("Loading plugins from config will ignore 'topic-check' section of config")

            plugins_config: list[Any] | dict[str, Any] = self.app_context.config.get("plugins", [])

            # if the config was generated from yaml, the plugins maybe a list instead of a dictionary; transform before loading
            #
            # plugins:
            #   - myproject.myfile.MyPlugin:

            if isinstance(plugins_config, list):
                plugins_info: dict[str, Any] = {}
                for plugin_config in plugins_config:
                    if isinstance(plugin_config, str):
                        plugins_info.update({plugin_config: {}})
                    elif not isinstance(plugin_config, dict):
                        msg = "malformed 'plugins' configuration"
                        raise PluginLoadError(msg)
                    else:
                        plugins_info.update(plugin_config)
                self._load_str_plugins(plugins_info)
            elif isinstance(plugins_config, dict):
                self._load_str_plugins(plugins_config)
        else:
            if not namespace:
                msg = "Namespace needs to be provided for EntryPoint plugin definitions"
                raise PluginLoadError(msg)

            warnings.warn(
                "Loading plugins from EntryPoints is deprecated and will be removed in a future version."
                " Use `plugins` section of config instead.",
                DeprecationWarning,
                stacklevel=4
            )

            self._load_ep_plugins(namespace)

        # for all the loaded plugins, find all event callbacks
        for plugin in self._plugins:
            for event in list(BrokerEvents) + list(MQTTEvents):
                if awaitable := getattr(plugin, f"on_{event}", None):
                    if not iscoroutinefunction(awaitable):
                        msg = f"'on_{event}' for '{plugin.__class__.__name__}' is not a coroutine'"
                        raise PluginImportError(msg)
                    self.logger.debug(f"'{event}' handler found for '{plugin.__class__.__name__}'")
                    self._event_plugin_callbacks[event].append(awaitable)

    def _load_ep_plugins(self, namespace: str) -> None:
        """Load plugins from `pyproject.toml` entrypoints. Deprecated."""
        self.logger.debug(f"Loading plugins for namespace {namespace}")
        auth_filter_list = []
        topic_filter_list = []
        if self.app_context.config and "auth" in self.app_context.config:
            auth_filter_list = self.app_context.config["auth"].get("plugins", None)
        if self.app_context.config and "topic-check" in self.app_context.config:
            topic_filter_list = self.app_context.config["topic-check"].get("plugins", None)

        ep: EntryPoints | list[EntryPoint] = []
        if hasattr(entry_points(), "select"):
            ep = entry_points().select(group=namespace)
        elif namespace in entry_points():
            ep = [entry_points()[namespace]]

        for item in ep:
            ep_plugin = self._load_ep_plugin(item)
            if ep_plugin is not None:
                self._plugins.append(ep_plugin.object)
                # maintain legacy behavior that if there is no list, use all auth plugins
                if ((auth_filter_list is None or ep_plugin.name in auth_filter_list)
                        and hasattr(ep_plugin.object, "authenticate")):
                    self._auth_plugins.append(ep_plugin.object)
                # maintain legacy behavior that if there is no list, use all topic plugins
                if ((topic_filter_list is None or ep_plugin.name in topic_filter_list)
                        and hasattr(ep_plugin.object, "topic_filtering")):
                    self._topic_plugins.append(ep_plugin.object)
                self.logger.debug(f" Plugin {item.name} ready")

    def _load_ep_plugin(self, ep: EntryPoint) -> Plugin | None:
        """Load plugins from `pyproject.toml` entrypoints. Deprecated."""
        try:
            self.logger.debug(f" Loading plugin {ep!s}")
            plugin = ep.load()

        except ImportError as e:
            self.logger.debug(f"Plugin import failed: {ep!r}", exc_info=True)
            raise PluginImportError(ep) from e

        self.logger.debug(f" Initializing plugin {ep!s}")

        plugin_context = copy.copy(self.app_context)
        plugin_context.logger = self.logger.getChild(ep.name)
        try:
            obj = plugin(plugin_context)
            return Plugin(ep.name, ep, obj)
        except Exception as e:
            self.logger.debug(f"Plugin init failed: {ep!r}", exc_info=True)
            raise PluginInitError(ep) from e

    def _load_str_plugins(self, plugins_info: dict[str, Any]) -> None:

        self.logger.info("Loading plugins from config")
        # legacy had a filtering 'enabled' flag, even if plugins were loaded/listed
        self._is_topic_filtering_enabled = True
        self._is_auth_filtering_enabled = True
        for plugin_path, plugin_config in plugins_info.items():

            plugin = self._load_str_plugin(plugin_path, plugin_config)
            self._plugins.append(plugin)

            # make sure that authenticate and topic filtering plugins have the appropriate async signature
            if isinstance(plugin, BaseAuthPlugin):
                if not iscoroutinefunction(plugin.authenticate):
                    msg = f"Auth plugin {plugin_path} has non-async authenticate method."
                    raise PluginCoroError(msg)
                self._auth_plugins.append(plugin)
            if isinstance(plugin, BaseTopicPlugin):
                if not iscoroutinefunction(plugin.topic_filtering):
                    msg = f"Topic plugin {plugin_path} has non-async topic_filtering method."
                    raise PluginCoroError(msg)
                self._topic_plugins.append(plugin)

    def _load_str_plugin(self, plugin_path: str, plugin_cfg: dict[str, Any] | None = None) -> "BasePlugin[C]":
        """Load plugin from string dotted path: mymodule.myfile.MyPlugin."""
        try:
            plugin_class: Any = import_string(plugin_path)
        except ImportError as ep:
            msg = f"Plugin import failed: {plugin_path}"
            raise PluginImportError(msg) from ep

        if not safe_issubclass(plugin_class, BasePlugin):
            msg = f"Plugin {plugin_path} is not a subclass of 'BasePlugin'"
            raise PluginLoadError(msg)

        plugin_context = copy.copy(self.app_context)
        plugin_context.logger = self.logger.getChild(plugin_class.__name__)
        try:
            # populate the config based on the inner dataclass called `Config`
            #   use `dacite` package to type check
            plugin_context.config = from_dict(data_class=plugin_class.Config,
                                              data=plugin_cfg or {},
                                              config=DaciteConfig(strict=True))
        except DaciteError as e:
            raise PluginLoadError from e
        except TypeError as e:
            msg = f"Could not marshall 'Config' of {plugin_path}; should be a dataclass."
            raise PluginLoadError(msg) from e

        try:
            pc = plugin_class(plugin_context)
            self.logger.debug(f"Loading plugin {plugin_path}")
            return cast("BasePlugin[C]", pc)
        except Exception as e:
            self.logger.debug(f"Plugin init failed: {plugin_class.__name__}", exc_info=True)
            raise PluginInitError(plugin_class) from e

    def get_plugin(self, name: str) -> Optional["BasePlugin[C]"]:
        """Get a plugin by its name from the plugins loaded for the current namespace.

        Only used for testing purposes to verify plugin loading correctly.

        :param name:
        :return:
        """
        for p in self._plugins:
            if p.__class__.__name__ == name:
                return p
        return None

    def is_topic_filtering_enabled(self) -> bool:
        topic_config = self.app_context.config.get("topic-check", {}) if self.app_context.config else {}
        if isinstance(topic_config, dict):
            return topic_config.get("enabled", False) or self._is_topic_filtering_enabled
        return False or self._is_topic_filtering_enabled

    async def close(self) -> None:
        """Free PluginManager resources and cancel pending event methods."""
        await self.map_plugin_close()
        for task in self._fired_events:
            task.cancel()
        self._fired_events.clear()

    @property
    def plugins(self) -> list["BasePlugin[C]"]:
        """Get the loaded plugins list.

        :return:
        """
        return self._plugins

    def _schedule_coro(self, coro: Awaitable[str | bool | None]) -> asyncio.Future[str | bool | None]:
        return asyncio.ensure_future(coro)

    def _clean_fired_events(self, future: asyncio.Future[Any]) -> None:
        if self.logger.getEffectiveLevel() <= logging.DEBUG:
            try:
                future.result()
            except asyncio.CancelledError:
                self.logger.warning("fired event was cancelled")
            # display plugin fault; don't allow it to cause a broker failure
            except Exception as exc:  # noqa: BLE001, pylint: disable=W0718
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)

        with contextlib.suppress(KeyError, ValueError):
            self._fired_events.remove(future)

    async def fire_event(self, event_name: Events, *, wait: bool = False, **method_kwargs: Any) -> None:
        """Fire an event to plugins.

        PluginManager schedules async calls for each plugin on method called "on_" + event_name.
        For example, on_connect will be called on event 'connect'.
        Method calls are scheduled in the async loop. wait parameter must be set to true
        to wait until all methods are completed.
        :param event_name:
        :param method_kwargs:
        :param wait: indicates if fire_event should wait for plugin calls completion (True), or not
        :return:
        """
        tasks: list[asyncio.Future[Any]] = []

        # check if any plugin has defined a callback for this event, skip if none
        if event_name not in self._event_plugin_callbacks:
            return

        for event_awaitable in self._event_plugin_callbacks[event_name]:

            async def call_method(method: AsyncFunc, kwargs: dict[str, Any]) -> Any:
                return await method(**kwargs)

            coro_instance: Awaitable[Any] = call_method(event_awaitable, method_kwargs)
            tasks.append(asyncio.ensure_future(coro_instance))
            tasks[-1].add_done_callback(self._clean_fired_events)

        self._fired_events.extend(tasks)
        if wait and tasks:
            await asyncio.wait(tasks)
        self.logger.debug(f"Plugins len(_fired_events)={len(self._fired_events)}")

    @staticmethod
    async def _map_plugin_method(
        plugins: list["BasePlugin[C]"],
        method_name: str,
        method_kwargs: dict[str, Any],
    ) -> dict["BasePlugin[C]", str | bool | None]:
        """Call plugin coroutines.

        :param plugins: List of plugins to execute the method on
        :param method_name: Name of the method to call on each plugin
        :param method_kwargs: Keyword arguments to pass to the method
        :return: dict containing return from coro call for each plugin.
        """
        tasks: list[asyncio.Future[Any]] = []

        for plugin in plugins:
            if not hasattr(plugin, method_name):
                continue

            async def call_method(p: "BasePlugin[C]", kwargs: dict[str, Any]) -> Any:
                method = getattr(p, method_name)
                return await method(**kwargs)

            coro_instance: Awaitable[Any] = call_method(plugin, method_kwargs)
            tasks.append(asyncio.ensure_future(coro_instance))

        ret_dict: dict[BasePlugin[C], str | bool | None] = {}
        if tasks:
            ret_list = await asyncio.gather(*tasks)
            ret_dict = dict(zip(plugins, ret_list, strict=False))

        return ret_dict

    async def map_plugin_auth(self, *, session: Session) -> dict["BasePlugin[C]", str | bool | None]:
        """Schedule a coroutine for plugin 'authenticate' calls.

        :param session: the client session associated with the authentication check
        :return: dict containing return from coro call for each plugin.
        """
        return await self._map_plugin_method(
            self._auth_plugins, "authenticate", {"session": session})  # type: ignore[arg-type]

    async def map_plugin_topic(
        self, *, session: Session, topic: str, action: "Action"
    ) -> dict["BasePlugin[C]", str | bool | None]:
        """Schedule a coroutine for plugin 'topic_filtering' calls.

        :param session: the client session associated with the topic_filtering check
        :param topic: the topic that needs to be filtered
        :param action: the action being executed
        :return: dict containing return from coro call for each plugin.
        """
        return await self._map_plugin_method(
            self._topic_plugins, "topic_filtering",   # type: ignore[arg-type]
            {"session": session, "topic": topic, "action": action}
        )

    async def map_plugin_close(self) -> None:
        """Schedule a coroutine for plugin 'close' calls.

        :return: dict containing return from coro call for each plugin.
        """
        await self._map_plugin_method(self._plugins, "close", {})



from dataclasses import dataclass, field
from pathlib import Path

from passlib.apps import custom_app_context as pwd_context

from amqtt.broker import BrokerContext
from amqtt.contexts import BaseContext
from amqtt.plugins.base import BaseAuthPlugin
from amqtt.session import Session

_PARTS_EXPECTED_LENGTH = 2  # Expected number of parts in a valid line


class AnonymousAuthPlugin(BaseAuthPlugin):
    """Authentication plugin allowing anonymous access."""

    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)

        # Default to allowing anonymous
        self._allow_anonymous = self._get_config_option("allow-anonymous", True)  # noqa: FBT003

    async def authenticate(self, *, session: Session) -> bool:
        authenticated = await super().authenticate(session=session)
        if authenticated:

            if self._allow_anonymous:
                self.context.logger.debug("Authentication success: config allows anonymous")
                session.is_anonymous = True
                return True

            if session and session.username:
                self.context.logger.debug(f"Authentication success: session has username '{session.username}'")
                return True
            self.context.logger.debug("Authentication failure: session has no username")
        return False

    @dataclass
    class Config:
        """Configuration for AnonymousAuthPlugin."""

        allow_anonymous: bool = field(default=True)
        """Allow all anonymous authentication (even with _no_ username)."""


class FileAuthPlugin(BaseAuthPlugin):
    """Authentication plugin based on a file-stored user database."""

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        self._users: dict[str, str] = {}
        self._read_password_file()

    def _read_password_file(self) -> None:
        """Read the password file and populates the user dictionary."""
        password_file = self._get_config_option("password-file", None)
        if not password_file:
            self.context.logger.warning("Configuration parameter 'password-file' not found")
            return

        try:
            file = password_file
            if isinstance(file, str):
                file = Path(file)
            with file.open(mode="r", encoding="utf-8") as file:
                self.context.logger.debug(f"Reading user database from {password_file}")
                for _line in file:
                    line = _line.strip()
                    if line and not line.startswith("#"):  # Skip empty lines and comments
                        parts = line.split(":", maxsplit=1)
                        if len(parts) == _PARTS_EXPECTED_LENGTH:
                            username, pwd_hash = parts
                            self._users[username] = pwd_hash
                            self.context.logger.debug(f"User '{username}' loaded")
                        else:
                            self.context.logger.warning(f"Malformed line in password file: {line}")
            self.context.logger.info(f"{len(self._users)} user(s) loaded from {password_file}")
        except FileNotFoundError:
            self.context.logger.warning(f"Password file '{password_file}' not found")
        except ValueError:
            self.context.logger.exception(f"Malformed password file '{password_file}'")
        except OSError:
            self.context.logger.exception(f"Unexpected error reading password file '{password_file}'")

    async def authenticate(self, *, session: Session) -> bool | None:
        """Authenticate users based on the file-stored user database."""
        authenticated = await super().authenticate(session=session)
        if authenticated:
            if not session:
                self.context.logger.debug("Authentication failure: no session provided")
                return False

            if not session.username:
                self.context.logger.debug("Authentication failure: no username provided in session")
                return None

            hash_session_username = self._users.get(session.username)
            if not hash_session_username:
                self.context.logger.debug(f"Authentication failure: no hash found for user '{session.username}'")
                return False

            if pwd_context.verify(session.password, hash_session_username):
                self.context.logger.debug(f"Authentication success for user '{session.username}'")
                return True

            self.context.logger.debug(f"Authentication failure: password mismatch for user '{session.username}'")
        return False

    @dataclass
    class Config:
        """Configuration for FileAuthPlugin."""

        password_file: str | Path | None = None
        """Path to file with `username:password` pairs, one per line. All passwords are encoded using sha-512."""



import ssl

import pytest

from amqtt.adapters import ReaderAdapter, WriterAdapter


class BrokenReaderAdapter(ReaderAdapter):

    async def read(self, n: int = -1) -> bytes:
        return await super().read(n)

    def feed_eof(self) -> None:
        return super().feed_eof()


@pytest.mark.asyncio
async def test_abstract_read_raises():
    reader = BrokenReaderAdapter()
    with pytest.raises(NotImplementedError):
        await reader.read()

    with pytest.raises(NotImplementedError):
        reader.feed_eof()

class BrokerWriterAdapter(WriterAdapter):
    def write(self, data: bytes) -> None:
        super().write(data)

    async def drain(self) -> None:
        await super().drain()

    def get_peer_info(self) -> tuple[str, int] | None:
        return super().get_peer_info()

    def get_ssl_info(self) -> ssl.SSLObject | None:
        return None

    async def close(self) -> None:
        await super().close()

@pytest.mark.asyncio
async def test_abstract_write_raises():
    writer = BrokerWriterAdapter()
    with pytest.raises(NotImplementedError):
        writer.write(b'')

    with pytest.raises(NotImplementedError):
        await writer.drain()

    with pytest.raises(NotImplementedError):
        writer.get_peer_info()

    with pytest.raises(NotImplementedError):
        await writer.close()



from collections import Counter
from collections.abc import MutableMapping
from dataclasses import dataclass, field

try:
    from enum import StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  # type: ignore[no-redef]
        pass
import time
from typing import Any, Generic, TypeVar

from mergedeep import merge

C = TypeVar("C", bound=Any)


class StateError(Exception):
    def __init__(self, msg: str = "'state' field is required") -> None:
        super().__init__(msg)


@dataclass
class MetaTimestamp:
    timestamp: int = 0

    def __eq__(self, other: object) -> bool:
        """Compare timestamps."""
        if isinstance(other, int):
            return self.timestamp == other
        if isinstance(other, self.__class__):
            return self.timestamp == other.timestamp
        msg = "needs to be int or MetaTimestamp"
        raise ValueError(msg)

    # numeric operations to make this dataclass transparent
    def __abs__(self) -> int:
        """Absolute timestamp."""
        return self.timestamp

    def __add__(self, other: int) -> int:
        """Add to a timestamp."""
        return self.timestamp + other

    def __sub__(self, other: int) -> int:
        """Subtract from a timestamp."""
        return self.timestamp - other

    def __mul__(self, other: int) -> int:
        """Multiply a timestamp."""
        return self.timestamp * other

    def __float__(self) -> float:
        """Convert timestamp to float."""
        return float(self.timestamp)

    def __int__(self) -> int:
        """Convert timestamp to int."""
        return int(self.timestamp)

    def __lt__(self, other: int) -> bool:
        """Compare timestamp."""
        return self.timestamp < other

    def __le__(self, other: int) -> bool:
        """Compare timestamp."""
        return self.timestamp <= other

    def __gt__(self, other: int) -> bool:
        """Compare timestamp."""
        return self.timestamp > other

    def __ge__(self, other: int) -> bool:
        """Compare timestamp."""
        return self.timestamp >= other


def create_metadata(state: MutableMapping[str, Any], timestamp: int) -> dict[str, Any]:
    """Create metadata (timestamps) for each of the keys in 'state'."""
    metadata: dict[str, Any] = {}
    for key, value in state.items():
        if isinstance(value, dict):
            metadata[key] = create_metadata(value, timestamp)
        elif value is None:
            metadata[key] = None
        else:
            metadata[key] = MetaTimestamp(timestamp)

    return metadata


def calculate_delta_update(desired: MutableMapping[str, Any],
                           reported: MutableMapping[str, Any],
                           depth: bool = True,
                           exclude_nones: bool = True,
                           ordered_lists: bool = True) -> dict[str, Any]:
    """Calculate state differences between desired and reported."""
    diff_dict = {}
    for key, value in desired.items():
        if value is None and exclude_nones:
            continue
        # if the desired has an element that the reported does not...
        if key not in reported:
            diff_dict[key] = value
        # if the desired has an element that's a list, but the list is
        elif isinstance(value, list) and not ordered_lists:
            if Counter(value) != Counter(reported[key]):
                diff_dict[key] = value
        elif isinstance(value, dict) and depth:
            # recurse, report when there is a difference
            obj_diff = calculate_delta_update(value, reported[key])
            if obj_diff:
                diff_dict[key] = obj_diff
        elif value != reported[key]:
            diff_dict[key] = value

    return diff_dict


def calculate_iota_update(desired: MutableMapping[str, Any], reported: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
    """Calculate state differences between desired and reported (including missing keys)."""
    delta = calculate_delta_update(desired, reported, depth=False, exclude_nones=False)

    for key in reported:
        if key not in desired:
            delta[key] = None

    return delta


@dataclass
class State(Generic[C]):
    desired: MutableMapping[str, C] = field(default_factory=dict)
    reported: MutableMapping[str, C] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict[str, C]) -> "State[C]":
        """Create state from dictionary."""
        return cls(
            desired=data.get("desired", {}),
            reported=data.get("reported", {})
        )

    def __bool__(self) -> bool:
        """Determine if state is empty."""
        return bool(self.desired) or bool(self.reported)

    def __add__(self, other: "State[C]") -> "State[C]":
        """Merge states together."""
        return State(
            desired=merge({}, self.desired, other.desired),
            reported=merge({}, self.reported, other.reported)
        )


@dataclass
class StateDocument:
    state: State[dict[str, Any]] = field(default_factory=State)
    metadata: State[MetaTimestamp] = field(default_factory=State)
    version: int | None = None  # only required when generating shadow messages
    timestamp: int | None = None  # only required when generating shadow messages

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "StateDocument":
        """Create state document from dictionary."""
        now = int(time.time())
        if data and "state" not in data:
            raise StateError

        state = State.from_dict(data.get("state", {}))
        metadata = State(
            desired=create_metadata(state.desired, now),
            reported=create_metadata(state.reported, now)
        )

        return cls(state=state, metadata=metadata)

    def __post_init__(self) -> None:
        """Initialize meta data if not provided."""
        now = int(time.time())
        if not self.metadata:
            self.metadata = State(
                desired=create_metadata(self.state.desired, now),
                reported=create_metadata(self.state.reported, now),
            )

    def __add__(self, other: "StateDocument") -> "StateDocument":
        """Merge two state documents together."""
        return StateDocument(
            state=self.state + other.state,
            metadata=self.metadata + other.metadata
        )


class ShadowOperation(StrEnum):
    GET = "get"
    UPDATE = "update"
    GET_ACCEPT = "get/accepted"
    GET_REJECT = "get/rejected"
    UPDATE_ACCEPT = "update/accepted"
    UPDATE_REJECT = "update/rejected"
    UPDATE_DOCUMENTS = "update/documents"
    UPDATE_DELTA = "update/delta"
    UPDATE_IOTA = "update/iota"



from dataclasses import dataclass
import logging
from pathlib import Path

from sqlalchemy import Boolean, Integer, LargeBinary, Result, String, select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from amqtt.broker import BrokerContext, RetainedApplicationMessage
from amqtt.contrib import DataClassListJSON
from amqtt.errors import PluginError
from amqtt.plugins.base import BasePlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


class Base(DeclarativeBase):
    pass


@dataclass
class RetainedMessage:
    topic: str
    data: str
    qos: int


@dataclass
class Subscription:
    topic: str
    qos: int


class StoredSession(Base):
    __tablename__ = "stored_sessions"

    id: Mapped[int] = mapped_column(primary_key=True)
    client_id: Mapped[str] = mapped_column(String)

    clean_session: Mapped[bool | None] = mapped_column(Boolean, nullable=True)

    will_flag: Mapped[bool] = mapped_column(Boolean, default=False, server_default="false")

    will_message: Mapped[bytes | None] = mapped_column(LargeBinary, nullable=True, default=None)
    will_qos: Mapped[int | None] = mapped_column(Integer, nullable=True, default=None)
    will_retain: Mapped[bool | None] = mapped_column(Boolean, nullable=True, default=None)
    will_topic: Mapped[str | None] = mapped_column(String, nullable=True, default=None)

    keep_alive: Mapped[int] = mapped_column(Integer, default=0)
    retained: Mapped[list[RetainedMessage]] = mapped_column(DataClassListJSON(RetainedMessage), default=list)
    subscriptions: Mapped[list[Subscription]] = mapped_column(DataClassListJSON(Subscription), default=list)


class StoredMessage(Base):
    __tablename__ = "stored_messages"

    id: Mapped[int] = mapped_column(primary_key=True)
    topic: Mapped[str] = mapped_column(String)
    data: Mapped[bytes | None] = mapped_column(LargeBinary, nullable=True, default=None)
    qos: Mapped[int] = mapped_column(Integer, default=0)


class SessionDBPlugin(BasePlugin[BrokerContext]):
    """Plugin to store session information and retained topic messages in the event that the broker terminates abnormally.

    Configuration:
        - file *(string)* path & filename to store the session db. default: `amqtt.db`
        - clear_on_shutdown *(bool)* if the broker shutdowns down normally, don't retain any information. default: `True`

    """

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        # bypass the `test_plugins_correct_has_attr` until it can be updated
        if not hasattr(self.config, "file"):
            logger.warning("`Config` is missing a `file` attribute")
            return

        self._engine = create_async_engine(f"sqlite+aiosqlite:///{self.config.file}")
        self._db_session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    @staticmethod
    async def _get_or_create_session(db_session: AsyncSession, client_id: str) -> StoredSession:

        stmt = select(StoredSession).filter(StoredSession.client_id == client_id)
        stored_session = await db_session.scalar(stmt)
        if stored_session is None:
            stored_session = StoredSession(client_id=client_id)
            db_session.add(stored_session)
        await db_session.flush()
        return stored_session

    @staticmethod
    async def _get_or_create_message(db_session: AsyncSession, topic: str) -> StoredMessage:

        stmt = select(StoredMessage).filter(StoredMessage.topic == topic)
        stored_message = await db_session.scalar(stmt)
        if stored_message is None:
            stored_message = StoredMessage(topic=topic)
            db_session.add(stored_message)
        await db_session.flush()
        return stored_message

    async def on_broker_client_connected(self, client_id: str, client_session: Session) -> None:
        """Search to see if session already exists."""
        # if client id doesn't exist, create (can ignore if session is anonymous)
        # update session information (will, clean_session, etc)

        # don't store session information for clean or anonymous sessions
        if client_session.clean_session in (None, True) or client_session.is_anonymous:
            return
        async with self._db_session_maker() as db_session, db_session.begin():
            stored_session = await self._get_or_create_session(db_session, client_id)

            stored_session.clean_session = client_session.clean_session
            stored_session.will_flag = client_session.will_flag
            stored_session.will_message = client_session.will_message  # type: ignore[assignment]
            stored_session.will_qos = client_session.will_qos
            stored_session.will_retain = client_session.will_retain
            stored_session.will_topic = client_session.will_topic
            stored_session.keep_alive = client_session.keep_alive

            await db_session.flush()

    async def on_broker_client_subscribed(self, client_id: str, topic: str, qos: int) -> None:
        """Create/update subscription if clean session = false."""
        session = self.context.get_session(client_id)
        if not session:
            logger.warning(f"'{client_id}' is subscribing but doesn't have a session")
            return

        if session.clean_session:
            return

        async with self._db_session_maker() as db_session, db_session.begin():
            # stored sessions shouldn't need to be created here, but we'll use the same helper...
            stored_session = await self._get_or_create_session(db_session, client_id)
            stored_session.subscriptions = [*stored_session.subscriptions, Subscription(topic, qos)]
            await db_session.flush()

    async def on_broker_client_unsubscribed(self, client_id: str, topic: str) -> None:
        """Remove subscription if clean session = false."""

    async def on_broker_retained_message(self, *, client_id: str | None, retained_message: RetainedApplicationMessage) -> None:
        """Update to retained messages.

        if retained_message.data is None or '', the message is being cleared
        """
        # if client_id is valid, the retained message is for a disconnected client
        if client_id is not None:
            async with self._db_session_maker() as db_session, db_session.begin():
                # stored sessions shouldn't need to be created here, but we'll use the same helper...
                stored_session = await self._get_or_create_session(db_session, client_id)
                stored_session.retained = [*stored_session.retained, RetainedMessage(retained_message.topic,
                                                                                     retained_message.data.decode(),
                                                                                     retained_message.qos or 0)]
                await db_session.flush()
            return

        async with self._db_session_maker() as db_session, db_session.begin():
            # if the retained message has data, we need to store/update for the topic
            if retained_message.data:
                client_message = await self._get_or_create_message(db_session, retained_message.topic)
                client_message.data = retained_message.data  # type: ignore[assignment]
                client_message.qos = retained_message.qos or 0
                await db_session.flush()
                return

            # if there is no data, clear the stored message (if exists) for the topic
            stmt = select(StoredMessage).filter(StoredMessage.topic == retained_message.topic)
            topic_message = await db_session.scalar(stmt)
            if topic_message is not None:
                await db_session.delete(topic_message)
                await db_session.flush()
            return

    async def on_broker_pre_start(self) -> None:
        """Initialize the database and db connection."""
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def on_broker_post_start(self) -> None:
        """Load subscriptions."""
        if len(self.context.subscriptions) > 0:
            msg = "SessionDBPlugin : broker shouldn't have any subscriptions yet"
            raise PluginError(msg)

        if len(list(self.context.sessions)) > 0:
            msg = "SessionDBPlugin : broker shouldn't have any sessions yet"
            raise PluginError(msg)

        async with self._db_session_maker() as db_session, db_session.begin():
            stmt = select(StoredSession)
            stored_sessions = await db_session.execute(stmt)

            restored_sessions = 0
            for stored_session in stored_sessions.scalars():
                await self.context.add_subscription(stored_session.client_id, None, None)
                for subscription in stored_session.subscriptions:
                    await self.context.add_subscription(stored_session.client_id,
                                                        subscription.topic,
                                                        subscription.qos)
                session = self.context.get_session(stored_session.client_id)
                if not session:
                    continue
                session.clean_session = stored_session.clean_session
                session.will_flag = stored_session.will_flag
                session.will_message = stored_session.will_message
                session.will_qos = stored_session.will_qos
                session.will_retain = stored_session.will_retain
                session.will_topic = stored_session.will_topic
                session.keep_alive = stored_session.keep_alive

                for message in stored_session.retained:
                    retained_message = RetainedApplicationMessage(
                        source_session=None,
                        topic=message.topic,
                        data=message.data.encode(),
                        qos=message.qos
                    )
                    await session.retained_messages.put(retained_message)
                restored_sessions += 1

            stmt = select(StoredMessage)
            stored_messages: Result[tuple[StoredMessage]] = await db_session.execute(stmt)

            restored_messages = 0
            retained_messages = self.context.retained_messages
            for stored_message in stored_messages.scalars():
                retained_messages[stored_message.topic] = (RetainedApplicationMessage(
                    source_session=None,
                    topic=stored_message.topic,
                    data=stored_message.data or b"",
                    qos=stored_message.qos
                ))
                restored_messages += 1
            logger.info(f"Retained messages restored: {restored_messages}")

        logger.info(f"Restored {restored_sessions} sessions.")

    async def on_broker_pre_shutdown(self) -> None:
        """Clean up the db connection."""
        await self._engine.dispose()

    async def on_broker_post_shutdown(self) -> None:

        if self.config.clear_on_shutdown and self.config.file.exists():
            self.config.file.unlink()

    @dataclass
    class Config:
        """Configuration variables."""

        file: str | Path = "amqtt.db"
        """path & filename to store the sqlite session db."""
        clear_on_shutdown: bool = True
        """if the broker shutdowns down normally, don't retain any information."""

        def __post_init__(self) -> None:
            """Create `Path` from string path."""
            if isinstance(self.file, str):
                self.file = Path(self.file)



import asyncio
import logging
import logging.config
import secrets
import socket
import string
import time
from unittest.mock import MagicMock, call, patch

import psutil
import pytest

from amqtt.events import BrokerEvents
from amqtt.adapters import StreamReaderAdapter, StreamWriterAdapter
from amqtt.broker import Broker
from amqtt.client import MQTTClient
from amqtt.errors import ConnectError
from amqtt.mqtt.connack import ConnackPacket
from amqtt.mqtt.connect import ConnectPacket, ConnectPayload, ConnectVariableHeader
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.disconnect import DisconnectPacket
from amqtt.mqtt.protocol.broker_handler import BrokerProtocolHandler
from amqtt.mqtt.pubcomp import PubcompPacket
from amqtt.mqtt.publish import PublishPacket
from amqtt.mqtt.pubrec import PubrecPacket
from amqtt.mqtt.pubrel import PubrelPacket
from amqtt.session import OutgoingApplicationMessage, Session

# formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
# logging.basicConfig(level=logging.DEBUG, format=formatter)

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,

    'formatters': {
        'default': {
            'format': '[%(asctime)s] %(levelname)s %(name)s: %(message)s',
        },
    },

    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'default',
            'stream': 'ext://sys.stdout',
        }
    },

    'root': {
        'handlers': ['console'],
        'level': 'DEBUG',
    },

    'loggers': {
        'transitions': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
    },
}

logging.config.dictConfig(LOGGING_CONFIG)






log = logging.getLogger(__name__)


# monkey patch MagicMock
# taken from https://stackoverflow.com/questions/51394411/python-object-magicmock-cant-be-used-in-await-expression
async def async_magic():
    pass


MagicMock.__await__ = lambda _: async_magic().__await__()


@pytest.mark.parametrize(
    "input_str, output_addr, output_port",
    [
        ("1234", None, 1234),
        (":1234", None, 1234),
        ("0.0.0.0:1234", "0.0.0.0", 1234),
        ("[::]:1234", "[::]", 1234),
        ("0.0.0.0", "0.0.0.0", 5678),
        ("[::]", "[::]", 5678),
        ("localhost", "localhost", 5678),
        ("localhost:1234", "localhost", 1234),
    ],
)
def test_split_bindaddr_port(input_str, output_addr, output_port):
    assert Broker._split_bindaddr_port(input_str, 5678) == (output_addr, output_port)


@pytest.mark.asyncio
async def test_start_stop(broker, mock_plugin_manager):
    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(BrokerEvents.PRE_START),
            call().fire_event(BrokerEvents.POST_START),
        ],
        any_order=True,
    )
    mock_plugin_manager.reset_mock()
    await broker.shutdown()
    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(BrokerEvents.PRE_SHUTDOWN),
            call().fire_event(BrokerEvents.POST_SHUTDOWN),
        ],
        any_order=True,
    )
    assert broker.transitions.is_stopped()


@pytest.mark.asyncio
async def test_client_connect(broker, mock_plugin_manager):
    client = MQTTClient(config={'auto_reconnect':False})

    ret = await client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    assert client.session is not None
    assert client.session.client_id in broker._sessions
    await client.disconnect()

    await asyncio.sleep(0.01)

    broker.plugins_manager.fire_event.assert_called()
    assert broker.plugins_manager.fire_event.call_count > 2

    # double indexing is ugly, but call_args_list returns a tuple of tuples
    events = [c[0][0] for c in broker.plugins_manager.fire_event.call_args_list]
    assert BrokerEvents.CLIENT_CONNECTED in events
    assert BrokerEvents.CLIENT_DISCONNECTED in events


@pytest.mark.asyncio
async def _connect_tcp(broker):
    process = psutil.Process()
    connections_number = 10

    # mqtt 3.1 requires a connect packet, otherwise the socket connection is rejected

    sockets = []
    for i in range(connections_number):
        static_connect_packet = b'\x10\x1b\x00\x04MQTT\x04\x02\x00<\x00\x0ftest-client-12' + f"{i}".encode()

        s = socket.create_connection(("127.0.0.1", 1883))
        s.send(static_connect_packet)
        sockets.append(s)

    # Wait for a brief moment to ensure connections are established
    await asyncio.sleep(0.1)

    # # Get the current number of TCP connections
    connections = process.net_connections()

    # max number of connections on the TCP listener is 10
    assert broker._servers["default"].conn_count == connections_number

    # Ensure connections are only on the TCP listener (port 1883)
    tcp_connections = [conn for conn in connections if conn.laddr.port == 1883]
    assert len(tcp_connections) == connections_number + 1  # Including the Broker's listening socket

    await asyncio.sleep(0.1)
    for conn in connections:
        assert conn.status in ("ESTABLISHED", "LISTEN")

    await asyncio.sleep(0.1)
    # close all connections
    for s in sockets:
        s.close()

    # Wait a moment for connections to be closed
    await asyncio.sleep(0.1)

    # Recheck connections after closing
    connections = process.net_connections()
    tcp_connections = [conn for conn in connections if conn.laddr.port == 1883]

    for conn in tcp_connections:
        assert conn.status in ("CLOSE_WAIT", "LISTEN")

    # Ensure no active connections for the default listener
    assert broker._servers["default"].conn_count == 0

    # Add one more connection to the TCP listener
    s = socket.create_connection(("127.0.0.1", 1883))
    s.send(static_connect_packet)

    open_connections = []
    open_connections = [conn for conn in process.net_connections() if conn.status == "ESTABLISHED"]

    # Ensure that only one TCP connection is active now
    assert len(open_connections) == 1
    await asyncio.sleep(0.1)
    assert broker._servers["default"].conn_count == 1


@pytest.mark.asyncio
async def test_client_connect_will_flag(broker):
    conn_reader, conn_writer = await asyncio.open_connection("127.0.0.1", 1883)
    reader = StreamReaderAdapter(conn_reader)
    writer = StreamWriterAdapter(conn_writer)

    vh = ConnectVariableHeader()
    payload = ConnectPayload()

    vh.keep_alive = 10
    vh.clean_session_flag = False
    vh.will_retain_flag = False
    vh.will_flag = True
    vh.will_qos = QOS_0
    payload.client_id = "test_id"
    payload.will_message = b"test"
    payload.will_topic = "/topic"
    connect = ConnectPacket(variable_header=vh, payload=payload)
    await connect.to_stream(writer)
    await ConnackPacket.from_stream(reader)

    await asyncio.sleep(0.1)

    disconnect = DisconnectPacket()
    await disconnect.to_stream(writer)

    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_connect_clean_session_false(broker):
    client = MQTTClient(client_id="", config={"auto_reconnect": False})
    return_code = None
    try:
        await client.connect("mqtt://127.0.0.1", cleansession=False)
    except ConnectError as ce:
        return_code = ce.return_code
    assert return_code == 0x02
    assert client.session is not None
    assert client.session.client_id not in broker._sessions
    await client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
@pytest.mark.parametrize('topic', [
    "/topic",
    "sport/#",
    "#",
    "sport/+",
    "sport/+/player",
])
async def test_client_subscribe(broker, mock_plugin_manager, topic):
    client = MQTTClient()
    ret = await client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    assert await client.subscribe([(topic, QOS_0)]) == [0x0]

    # Test if the client test client subscription is registered
    assert topic in broker._subscriptions
    subs = broker._subscriptions[topic]
    assert len(subs) == 1
    (s, qos) = subs[0]
    assert s == client.session
    assert qos == QOS_0

    await client.disconnect()
    await asyncio.sleep(0.1)


    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(
                BrokerEvents.CLIENT_SUBSCRIBED,
                client_id=client.session.client_id,
                topic=topic,
                qos=QOS_0,
            ),
        ],
        any_order=True,
    )


def assert_not_called_with_param(mock_obj, param_name=None, param_value=None):
    """Asserts a specific param name or value was never passed to the mock."""
    for args, kwargs in mock_obj.call_args_list:
        # Check keyword arguments
        if param_name in kwargs and kwargs[param_name] == param_value:
            raise AssertionError(f"Mock was called with unexpected kwarg {param_name}={param_value}")

        # Check positional arguments
        if param_value in args and param_name is None:
            raise AssertionError(f"Mock was called with unexpected positional arg {param_value}")


@pytest.mark.asyncio
@pytest.mark.parametrize('topic', [
    "sport/#/player",
    "sport/foo#",
    "sport/##",
    "sport/foo+bar",
    "sport/++"
])
async def test_client_fails_subscribe(broker, mock_plugin_manager, topic):
    client = MQTTClient()
    ret = await client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    assert await client.subscribe([(topic, QOS_0)]) == [0x80]

    assert topic not in broker._subscriptions

    await client.disconnect()
    await asyncio.sleep(0.1)

    assert_not_called_with_param(mock_plugin_manager, param_value=BrokerEvents.CLIENT_SUBSCRIBED)

@pytest.mark.asyncio
async def test_client_subscribe_twice(broker, mock_plugin_manager):
    client = MQTTClient()
    ret = await client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    await client.subscribe([("/topic", QOS_0)])

    # Test if the client test client subscription is registered
    assert "/topic" in broker._subscriptions
    subs = broker._subscriptions["/topic"]
    assert len(subs) == 1
    (s, qos) = subs[0]
    assert s == client.session
    assert qos == QOS_0

    await client.subscribe([("/topic", QOS_0)])
    assert len(subs) == 1
    (s, qos) = subs[0]
    assert s == client.session
    assert qos == QOS_0

    await client.disconnect()
    await asyncio.sleep(0.1)

    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(
                BrokerEvents.CLIENT_SUBSCRIBED,
                client_id=client.session.client_id,
                topic="/topic",
                qos=QOS_0,
            ),
        ],
        any_order=True,
    )


@pytest.mark.asyncio
async def test_client_unsubscribe(broker, mock_plugin_manager):
    client = MQTTClient()
    ret = await client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    await client.subscribe([("/topic", QOS_0)])

    # Test if the client test client subscription is registered
    assert "/topic" in broker._subscriptions
    subs = broker._subscriptions["/topic"]
    assert len(subs) == 1
    (s, qos) = subs[0]
    assert s == client.session
    assert qos == QOS_0

    await client.unsubscribe(["/topic"])
    await asyncio.sleep(0.1)
    assert broker._subscriptions["/topic"] == []
    await client.disconnect()
    await asyncio.sleep(0.1)

    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(
                BrokerEvents.CLIENT_SUBSCRIBED,
                client_id=client.session.client_id,
                topic="/topic",
                qos=QOS_0,
            ),
            call().fire_event(
                BrokerEvents.CLIENT_UNSUBSCRIBED,
                client_id=client.session.client_id,
                topic="/topic",
            ),
        ],
        any_order=True,
    )


@pytest.mark.asyncio
async def test_client_publish(broker, mock_plugin_manager):
    pub_client = MQTTClient()
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0

    ret_message = await pub_client.publish("/topic", b"data", QOS_0)
    await pub_client.disconnect()
    assert broker._retained_messages == {}

    await asyncio.sleep(0.1)
    assert pub_client.session is not None

    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(
                BrokerEvents.MESSAGE_RECEIVED,
                client_id=pub_client.session.client_id,
                message=ret_message,
            ),
        ],
        any_order=True,
    )


@pytest.mark.asyncio
async def test_client_publish_acl_permitted(acl_broker):
    sub_client = MQTTClient()
    ret_conn = await sub_client.connect("mqtt://user2:user2password@127.0.0.1:1884/")
    assert ret_conn == 0

    ret_sub = await sub_client.subscribe([("public/subtopic/test", QOS_0)])
    assert ret_sub == [QOS_0]

    pub_client = MQTTClient()
    ret_conn = await pub_client.connect("mqtt://user1:user1password@127.0.0.1:1884/")
    assert ret_conn == 0

    await pub_client.publish("public/subtopic/test", b"data", QOS_0)

    message = await sub_client.deliver_message(timeout_duration=1)
    await pub_client.disconnect()
    await sub_client.disconnect()

    assert message is not None
    assert message.topic == "public/subtopic/test"
    assert message.data == b"data"
    assert message.qos == QOS_0


@pytest.mark.asyncio
async def test_client_publish_acl_forbidden(acl_broker):
    sub_client = MQTTClient()
    ret_conn = await sub_client.connect("mqtt://user2:user2password@127.0.0.1:1884/")
    assert ret_conn == 0

    ret_sub = await sub_client.subscribe([("public/forbidden/test", QOS_0)])
    assert ret_sub == [QOS_0]

    pub_client = MQTTClient()
    ret_conn = await pub_client.connect("mqtt://user1:user1password@127.0.0.1:1884/")
    assert ret_conn == 0

    await pub_client.publish("public/forbidden/test", b"data", QOS_0)

    try:
        await sub_client.deliver_message(timeout_duration=1)
        msg = "Should not have worked"
        raise AssertionError(msg)
    except Exception:
        pass

    await pub_client.disconnect()
    await sub_client.disconnect()


@pytest.mark.asyncio
async def test_client_publish_acl_permitted_sub_forbidden(acl_broker):
    sub_client1 = MQTTClient(client_id="sub_client1")
    ret_conn = await sub_client1.connect("mqtt://user2:user2password@127.0.0.1:1884/")
    assert ret_conn == 0

    sub_client2 = MQTTClient(client_id="sub_client2")
    ret_conn = await sub_client2.connect("mqtt://user3:user3password@127.0.0.1:1884/")
    assert ret_conn == 0

    ret_sub = await sub_client1.subscribe([("public/subtopic/test", QOS_0)])
    assert ret_sub == [QOS_0]

    ret_sub = await sub_client2.subscribe([("public/subtopic/test", QOS_0)])
    assert ret_sub == [128]

    pub_client = MQTTClient(client_id="pub_client")
    ret_conn = await pub_client.connect("mqtt://user1:user1password@127.0.0.1:1884/")
    assert ret_conn == 0

    await pub_client.publish("public/subtopic/test", b"data", QOS_0)

    message = await sub_client1.deliver_message(timeout_duration=1)

    try:
        await sub_client2.deliver_message(timeout_duration=1)
        msg = "Should not have worked"
        raise AssertionError(msg)
    except Exception:
        pass

    await pub_client.disconnect()
    await sub_client1.disconnect()
    await sub_client2.disconnect()

    assert message is not None
    assert message.topic == "public/subtopic/test"
    assert message.data == b"data"
    assert message.qos == QOS_0


@pytest.mark.asyncio
async def test_client_publish_dup(broker):
    conn_reader, conn_writer = await asyncio.open_connection("127.0.0.1", 1883)
    reader = StreamReaderAdapter(conn_reader)
    writer = StreamWriterAdapter(conn_writer)

    vh = ConnectVariableHeader()
    payload = ConnectPayload()

    vh.keep_alive = 10
    vh.clean_session_flag = False
    vh.will_retain_flag = False
    payload.client_id = "test_id"
    connect = ConnectPacket(variable_header=vh, payload=payload)
    await connect.to_stream(writer)
    await ConnackPacket.from_stream(reader)

    publish_1 = PublishPacket.build("/test", b"data", 1, False, QOS_2, False)
    await publish_1.to_stream(writer)

    # Store the future of PubrecPacket.from_stream() in a variable
    pubrec_task_1 = asyncio.ensure_future(PubrecPacket.from_stream(reader))

    await asyncio.sleep(2)

    publish_dup = PublishPacket.build("/test", b"data", 1, True, QOS_2, False)
    await publish_dup.to_stream(writer)
    await PubrecPacket.from_stream(reader)
    pubrel = PubrelPacket.build(1)
    await pubrel.to_stream(writer)
    await PubcompPacket.from_stream(reader)

    # Ensure we wait for the Pubrec packets to be processed
    await pubrec_task_1

    disconnect = DisconnectPacket()
    await disconnect.to_stream(writer)


@pytest.mark.asyncio
async def test_client_publishing_invalid_topic(broker):
    assert broker.transitions.is_started()
    pub_client = MQTTClient(config={'auto_reconnect': False})
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0

    await pub_client.subscribe([
        ("my/+/topic", QOS_0)
    ])
    await asyncio.sleep(0.5)

    # need to build & send packet directly to bypass client's check of invalid topic name
    #    see test_client.py::test_publish_to_incorrect_wildcard for client checks
    packet = PublishPacket.build(topic_name='my/topic', message=b'messages',
                                 packet_id=None, dup_flag=False, qos=QOS_0, retain=False)
    packet.topic_name = "my/+/topic"
    await pub_client._handler._send_packet(packet)
    await asyncio.sleep(0.5)

    with pytest.raises(asyncio.TimeoutError):
        msg = await pub_client.deliver_message(timeout_duration=1)
        assert msg is None

    await asyncio.sleep(0.1)
    await pub_client.disconnect()


@pytest.mark.asyncio
async def test_client_publish_asterisk(broker):
    """'*' is a valid, non-wildcard character for MQTT."""
    assert broker.transitions.is_started()
    pub_client = MQTTClient(config={'auto_reconnect': False})
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0

    await pub_client.subscribe([
        ("my*/topic", QOS_0),
        ("my/+/topic", QOS_0)
    ])
    await asyncio.sleep(0.1)
    await pub_client.publish('my*/topic', b'my valid message', QOS_0, retain=False)
    await asyncio.sleep(0.1)
    msg = await pub_client.deliver_message(timeout_duration=1)
    assert msg is not None
    assert msg.topic == "my*/topic"
    assert msg.data == b'my valid message'

    await asyncio.sleep(0.1)
    msg = await pub_client.publish('my/****/topic', b'my valid message', QOS_0, retain=False)
    assert msg is not None
    assert msg.topic == "my/****/topic"
    assert msg.data == b'my valid message'

    await pub_client.disconnect()


@pytest.mark.asyncio
async def test_client_publish_big(broker, mock_plugin_manager):
    pub_client = MQTTClient()
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0

    ret_message = await pub_client.publish(
        "/topic",
        bytearray(b"\x99" * 256 * 1024),
        QOS_2,
    )
    await pub_client.disconnect()
    assert broker._retained_messages == {}

    await asyncio.sleep(0.1)
    assert pub_client.session is not None

    mock_plugin_manager.assert_has_calls(
        [
            call().fire_event(
                BrokerEvents.MESSAGE_RECEIVED,
                client_id=pub_client.session.client_id,
                message=ret_message,
            ),
        ],
        any_order=True,
    )


@pytest.mark.asyncio
async def test_client_publish_retain(broker):
    pub_client = MQTTClient()
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    await pub_client.publish("/topic", b"data", QOS_0, retain=True)
    await pub_client.disconnect()
    await asyncio.sleep(0.1)
    assert "/topic" in broker._retained_messages
    retained_message = broker._retained_messages["/topic"]
    assert retained_message.source_session == pub_client.session
    assert retained_message.topic == "/topic"
    assert retained_message.data == b"data"
    assert retained_message.qos == QOS_0


@pytest.mark.asyncio
async def test_client_publish_retain_delete(broker):
    pub_client = MQTTClient()
    ret = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    await pub_client.publish("/topic", b"", QOS_0, retain=True)
    await pub_client.disconnect()
    await asyncio.sleep(0.1)
    assert "/topic" not in broker._retained_messages


@pytest.mark.asyncio
async def test_client_subscribe_publish(broker):
    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe(
        [("/qos0", QOS_0), ("/qos1", QOS_1), ("/qos2", QOS_2)],
    )
    assert ret == [QOS_0, QOS_1, QOS_2]

    await _client_publish("/qos0", b"data", QOS_0)
    await _client_publish("/qos1", b"data", QOS_1)
    await _client_publish("/qos2", b"data", QOS_2)
    await asyncio.sleep(0.1)
    for qos in [QOS_0, QOS_1, QOS_2]:
        message = await sub_client.deliver_message()
        assert message is not None
        assert message.topic == f"/qos{qos}"
        assert message.data == b"data"
        assert message.qos == qos
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_subscribe_invalid(broker):
    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe(
        [
            ("+", QOS_0),
            ("+/tennis/#", QOS_0),
            ("sport+", QOS_0),
            ("sport/+/player1", QOS_0),
        ],
    )
    assert ret == [QOS_0, QOS_0, 0x80, QOS_0]

    await asyncio.sleep(0.1)
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_subscribe_publish_dollar_topic_1(broker):
    assert broker.transitions.is_started()
    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe([("#", QOS_0)])
    assert ret == [QOS_0]

    await _client_publish("/topic", b"data", QOS_0)
    message = await sub_client.deliver_message()
    assert message is not None

    await _client_publish("$topic", b"data", QOS_0)
    await asyncio.sleep(0.1)
    message = None
    try:
        message = await sub_client.deliver_message(timeout_duration=2)
    except Exception:
        pass
    except RuntimeError as e:
        # The loop is closed with pending tasks. Needs fine tuning.
        log.warning(e)
    assert message is None
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_subscribe_publish_dollar_topic_2(broker):
    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe([("+/monitor/Clients", QOS_0)])
    assert ret == [QOS_0]

    await _client_publish("test/monitor/Clients", b"data", QOS_0)
    message = await sub_client.deliver_message()
    assert message is not None

    await _client_publish("$SYS/monitor/Clients", b"data", QOS_0)
    await asyncio.sleep(0.1)
    message = None
    try:
        message = await sub_client.deliver_message(timeout_duration=2)
    except Exception:
        pass
    except RuntimeError as e:
        # The loop is closed with pending tasks. Needs fine tuning.
        log.warning(e)
    assert message is None
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_publish_clean_session_subscribe(broker):

    sub_client = MQTTClient(client_id='test_client', config={'auto_reconnect': False})
    await sub_client.connect("mqtt://127.0.0.1", cleansession=False)
    ret = await sub_client.subscribe(
        [("/qos0", QOS_0), ("/qos1", QOS_1), ("/qos2", QOS_2)],
    )
    assert ret == [QOS_0, QOS_1, QOS_2]

    await sub_client.disconnect()
    await asyncio.sleep(0.5)


    await _client_publish("/qos0", b"data0", QOS_0)  # should not be retained
    await _client_publish("/qos1", b"data1", QOS_1)
    await _client_publish("/qos2", b"data2", QOS_2)
    await asyncio.sleep(2)

    await sub_client.reconnect(cleansession=False)
    for qos in [QOS_1, QOS_2]:
        log.debug(f"TEST QOS: {qos}")
        message = await sub_client.deliver_message(timeout_duration=2)
        log.debug(f"Message: {message.publish_packet if message else None!r}")
        assert message is not None
        assert message.topic == f"/qos{qos}"
        assert message.data == f"data{qos}".encode("utf-8")
        assert message.qos == qos

    try:
        while True:
            message = await sub_client.deliver_message(timeout_duration=1)
            assert message is not None, "no other messages should have been retained"
    except asyncio.TimeoutError:
        pass

    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_publish_retain_with_new_subscribe(broker):
    await asyncio.sleep(2)
    sub_client1 = MQTTClient(client_id='test_client1')
    await sub_client1.connect("mqtt://127.0.0.1")

    await sub_client1.disconnect()
    await asyncio.sleep(0.5)

    await _client_publish("/qos0", b"data0", QOS_0, retain=True)
    await asyncio.sleep(0.5)

    sub_client2 = MQTTClient(client_id='test_client2')
    await sub_client2.connect("mqtt://127.0.0.1")

    # should receive the retained message on subscription
    ret = await sub_client2.subscribe(
        [("/qos0", QOS_0)],
    )
    assert ret == [QOS_0]

    message = await sub_client2.deliver_message(timeout_duration=1)
    assert message is not None
    assert message.topic == "/qos0"
    assert message.data == b"data0"
    assert message.qos == QOS_0
    await sub_client2.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_publish_retain_latest_with_new_subscribe(broker):
    await asyncio.sleep(2)
    sub_client1 = MQTTClient(client_id='test_client1')
    await sub_client1.connect("mqtt://127.0.0.1")

    await sub_client1.disconnect()
    await asyncio.sleep(0.5)

    await _client_publish("/qos0", b"data a", QOS_0, retain=True)
    await asyncio.sleep(0.5)

    sub_client2 = MQTTClient(client_id='test_client2')
    await sub_client2.connect("mqtt://127.0.0.1")

    await _client_publish("/qos0", b"data b", QOS_0, retain=True)

    # should receive the retained message on subscription
    ret = await sub_client2.subscribe(
        [("/qos0", QOS_0)],
    )
    assert ret == [QOS_0]

    message = await sub_client2.deliver_message(timeout_duration=1)
    assert message is not None
    assert message.topic == "/qos0"
    assert message.data == b"data b"
    assert message.qos == QOS_0
    await sub_client2.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_client_publish_retain_subscribe_on_reconnect(broker):
    await asyncio.sleep(2)
    sub_client = MQTTClient(client_id='test_client')
    await sub_client.connect("mqtt://127.0.0.1", cleansession=False)
    ret = await sub_client.subscribe(
        [("/qos0", QOS_0)],
    )
    assert ret == [QOS_0]

    await sub_client.disconnect()
    await asyncio.sleep(0.5)

    await _client_publish("/qos0", b"data0", QOS_0, retain=True)
    await asyncio.sleep(0.5)

    await sub_client.reconnect(cleansession=False)

    message = await sub_client.deliver_message(timeout_duration=1)
    assert message is not None
    assert message.topic == "/qos0"
    assert message.data == b"data0"
    assert message.qos == QOS_0
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def _client_publish(topic, data, qos, retain=False) -> int | OutgoingApplicationMessage:

    gen_id = "pub_"
    valid_chars = string.ascii_letters + string.digits
    gen_id += "".join(secrets.choice(valid_chars) for _ in range(16))

    pub_client = MQTTClient(client_id=gen_id)
    ret: int | OutgoingApplicationMessage = await pub_client.connect("mqtt://127.0.0.1/")
    assert ret == 0
    ret = await pub_client.publish(topic, data, qos, retain)
    await pub_client.disconnect()
    return ret


def test_matches_multi_level_wildcard(broker):
    test_filter = "sport/tennis/player1/#"

    for bad_topic in [
        "sport/tennis",
        "sport/tennis/",
    ]:
        assert not broker._matches(bad_topic, test_filter)

    for good_topic in [
        "sport/tennis/player1",
        "sport/tennis/player1/",
        "sport/tennis/player1/ranking",
        "sport/tennis/player1/score/wimbledon",
    ]:
        assert broker._matches(good_topic, test_filter)


def test_matches_single_level_wildcard(broker):
    test_filter = "sport/tennis/+"

    for bad_topic in [
        "sport/tennis",
        "sport/tennis/player1/",
        "sport/tennis/player1/ranking",
    ]:
        assert not broker._matches(bad_topic, test_filter)

    for good_topic in [
        "sport/tennis/",
        "sport/tennis/player1",
        "sport/tennis/player2",
    ]:
        assert broker._matches(good_topic, test_filter)


def test_matches_plus_wildcard_redos_protection(broker):
    invalid_filter = "+/" + "+" * 60  # accepted by the broken validation, ReDoS in _matches
    trigger_topic = "a/" + "b" * 12 + "/x"  # legal topic that makes the filter backtrack forever

    start_time = time.perf_counter()
    # It must return False, but it must also do it instantly
    assert not broker._matches(trigger_topic, invalid_filter)
    duration = time.perf_counter() - start_time

    # If a regex backtracks exponentially, it takes seconds/minutes
    #   pytest's timeout should catch this failure; also, do an explicit check
    # An iterative or healthy parser will finish in less than 1 millisecond.
    assert duration < 0.01, f"Possible ReDoS detected! Execution took {duration:.4f}s"


@pytest.mark.parametrize(
    ("subscription_filter", "topic", "expected"),
    [
        # Exact matching
        ("sport/tennis", "sport/tennis", True),
        ("sport/tennis", "sport/football", False),
        ("sport/tennis", "sport/tennis/player1", False),
        ("sport/tennis/player1", "sport/tennis", False),

        # Valid single-level wildcard
        ("sport/+", "sport/tennis", True),
        ("sport/+/player1", "sport/tennis/player1", True),
        ("sport/+", "sport/tennis/player1", False),
        ("+", "sport", True),

        # '+' may match an empty level
        ("sport/+", "sport/", True),
        ("sport/+/player1", "sport//player1", True),

        # Valid multi-level wildcard
        ("#", "sport/tennis/player1", True),
        ("sport/#", "sport/tennis/player1", True),
        ("sport/#", "sport", True),
        ("sport/#", "sports", False),

        # Embedded '+' is invalid
        ("sport/tennis+", "sport/tennis", False),
        ("sport/tennis+", "sport/tennis1", False),
        ("foo+bar", "foo/bar", False),
        ("foo+bar", "foobeefbeefbar", False),
        ("sport/++++", "sport/tennis", False),

        # Embedded or misplaced '#' is invalid
        ("foo#", "foobar", False),
        ("#foo", "anything", False),
        ("foo#bar", "fooanythingbar", False),
        ("sport/####", "sport/tennis/player1", False),
        ("sport/#/player1", "sport/tennis/player1", False),
        ("sport/tennis#", "sport/tennis", False),

        # Mixed malformed wildcards
        ("sport/+#", "sport/tennis", False),
        ("sport/#+", "sport/tennis", False),

        # Original pathological pattern
        ("+/+" + "+" * 60, "a/b/c", False),
    ],
)
def test_matches_fails_safely_on_all_invalid_filters(broker, subscription_filter, topic, expected):
    # Defensive programming: even if a bad subscription leaks into the router,
    # it must safely evaluate to False instead of throwing exceptions or false positives.
    start_time = time.perf_counter()
    assert broker._matches(topic, subscription_filter) is expected
    duration = time.perf_counter() - start_time
    assert duration < 0.01, f"Possible ReDoS detected! Execution took {duration:.4f}s"


@pytest.mark.asyncio
@pytest.mark.parametrize(('topic', 'expected'), [
    ("sport/#", True),
    ("#", True),
    ("sport/+", True),
    ("sport/+/player", True),
    ("sport/#/player", False),
    ("sport/foo#", False),
    ("sport/##", False),
    ("sport/foo+bar", False),
    ("sport/++", False),
])
async def test_invalid_wildcard_subscriptions(broker, topic, expected):

    session = Session()
    assert (await broker.add_subscription((topic, 0), session) != 0x80) is expected


@pytest.mark.asyncio
async def test_broker_broadcast_cancellation(broker):
    topic = "test"
    data = b"data"
    qos = QOS_0

    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    await sub_client.subscribe([(topic, qos)])

    with patch.object(BrokerProtocolHandler, "mqtt_publish", side_effect=asyncio.CancelledError) as mocked_mqtt_publish:
        await _client_publish(topic, data, qos)

        # Second publish triggers the awaiting of first `mqtt_publish` task
        await _client_publish(topic, data, qos)
        await asyncio.sleep(0.01)

        mocked_mqtt_publish.assert_awaited()

    # Ensure broadcast loop is still functional and can deliver the message
    await _client_publish(topic, data, qos)
    message = await asyncio.wait_for(sub_client.deliver_message(), timeout=1)
    assert message


@pytest.mark.asyncio
async def test_broker_socket_open_close(broker):

    # check that https://github.com/Yakifo/amqtt/issues/86 is fixed

    # mqtt 3.1 requires a connect packet, otherwise the socket connection is rejected
    static_connect_packet = b'\x10\x1b\x00\x04MQTT\x04\x02\x00<\x00\x0ftest-client-123'
    s = socket.create_connection(("127.0.0.1", 1883))
    s.send(static_connect_packet)
    await asyncio.sleep(0.1)
    s.close()

std_legacy_config = {
        "listeners": {
            "default": {
                "type": "tcp",
                "bind": f"127.0.0.1:1883",
            }
        },
        "sys_interval": 10,
        "auth": {
            "allow-anonymous": True,
            "plugins": ["auth_anonymous"],
        },
        "topic-check": {"enabled": False},
    }


@pytest.mark.asyncio
async def test_broker_with_legacy_config():

    broker = Broker(config=std_legacy_config)

    await broker.start()
    await asyncio.sleep(2)

    mqtt_client = MQTTClient(config={'auto_reconnect': False})
    await mqtt_client.connect()

    await broker.shutdown()


legacy_config_empty_auth_plugin_list = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 0,
        'auth':{
            'plugins':[]  # explicitly declare no auth plugins
        }
    }
class_path_config_no_auth = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'plugins':{
            'tests.plugins.test_plugins.AllEventsPlugin': {}
        }
    }


@pytest.mark.parametrize("test_config", [
    legacy_config_empty_auth_plugin_list,
    class_path_config_no_auth,
])
@pytest.mark.asyncio
async def test_broker_without_auth_plugin(test_config):

    broker = Broker(config=test_config)

    await broker.start()
    await asyncio.sleep(2)

    # make sure all expected events get triggered
    with pytest.raises(ConnectError):
        mqtt_client = MQTTClient(config={'auto_reconnect': False})
        await mqtt_client.connect()


    await broker.shutdown()



legacy_config_with_absent_auth_plugin_filter = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 0,
        'auth':{
            'allow-anonymous': True
        }
    }
@pytest.mark.asyncio
async def test_broker_with_absent_auth_plugin_filter():

    # maintain legacy behavior that if a config is missing the 'auth' > 'plugins' filter, all plugins are active
    broker = Broker(config=legacy_config_with_absent_auth_plugin_filter)

    await broker.start()
    await asyncio.sleep(2)


    mqtt_client = MQTTClient(config={'auto_reconnect': False})
    await mqtt_client.connect()

    await broker.shutdown()



import asyncio
import logging

import pytest

from amqtt.broker import Broker
from amqtt.client import MQTTClient
from amqtt.contexts import BrokerConfig, ListenerConfig, ConnectionConfig, ClientConfig

logger = logging.getLogger(__name__)

@pytest.fixture
def session_broker_config():
     return BrokerConfig(
        listeners={
            'default': ListenerConfig(bind='127.0.0.1:1883')
        },
        session_expiry_interval=2,
        plugins={
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {'allow_anonymous': False}
        })


@pytest.mark.parametrize("username,clean_session,expiration,session_count",
                         [
                             # session expiration disabled
                             ("", True, None, 0),  # anonymous and clean session
                             ("", False, None, 0), # anonymous
                             ("myuser@", True, None, 0), # named user, clean session
                             ("myuser@", False, None, 1), # named user

                             # session expiration enabled
                             ("myuser@", False, 1, 0), # named user, quick expiration
                             ("myuser@", False, 20, 1), # named user, long expiration
                         ])
@pytest.mark.asyncio
async def test_clear_session_expiration(caplog, session_broker_config, username, clean_session, session_count, expiration):
    caplog.set_level(logging.DEBUG)

    session_broker_config.session_expiry_interval = expiration
    session_broker_config.plugins = {'amqtt.plugins.authentication.AnonymousAuthPlugin': {'allow_anonymous': username == ""}}

    broker = Broker(config=session_broker_config)
    await broker.start()
    await asyncio.sleep(0.1)
    assert len(broker._sessions) < 1

    c = MQTTClient(config=ClientConfig(cleansession=clean_session, auto_reconnect=False))
    await c.connect(f'mqtt://{username}127.0.0.1:1883')
    await asyncio.sleep(0.1)
    assert len(broker._sessions) == 1, "client should be connected"
    await asyncio.sleep(0.1)
    await c.disconnect()
    await asyncio.sleep(4)
    assert len(broker._sessions) == session_count, f"session counts don't match {len(broker._sessions)} v {session_count}"

    if not session_count:
        assert any([record for record in caplog.records if "Expired 1 sessions" in record.message]) == (session_count == 0)

    await broker.shutdown()



"""Module for contributed plugins."""

from dataclasses import asdict, is_dataclass
from typing import Any, TypeVar

from sqlalchemy import JSON, TypeDecorator

T = TypeVar("T")


class DataClassListJSON(TypeDecorator[list[dict[str, Any]]]):
    impl = JSON
    cache_ok = True

    def __init__(self, dataclass_type: type[T]) -> None:
        if not is_dataclass(dataclass_type):
            msg = f"{dataclass_type} must be a dataclass type"
            raise TypeError(msg)
        self.dataclass_type = dataclass_type
        super().__init__()

    def process_bind_param(
            self,
            value: list[Any] | None,  # Python -> DB
            dialect: Any
    ) -> list[dict[str, Any]] | None:
        if value is None:
            return None
        return [asdict(item) for item in value]

    def process_result_value(
            self,
            value: list[dict[str, Any]] | None,  # DB -> Python
            dialect: Any
    ) -> list[Any] | None:
        if value is None:
            return None
        return [self.dataclass_type(**item) for item in value]

    def process_literal_param(self, value: Any, dialect: Any) -> Any:
        # Required by SQLAlchemy, typically used for literal SQL rendering.
        return value

    @property
    def python_type(self) -> type:
        # Required by TypeEngine to indicate the expected Python type.
        return list



import asyncio
import sqlite3
import tempfile
from pathlib import Path

import pytest
import aiosqlite
from passlib.context import CryptContext

from amqtt.broker import BrokerContext, Broker
from amqtt.client import MQTTClient
from amqtt.contexts import Action
from amqtt.contrib.auth_db.models import AllowedTopic, PasswordHasher
from amqtt.contrib.auth_db.plugin import UserAuthDBPlugin, TopicAuthDBPlugin
from amqtt.contrib.auth_db.managers import UserManager, TopicManager
from amqtt.errors import ConnectError, MQTTError
from amqtt.mqtt.constants import QOS_1, QOS_0
from amqtt.session import Session
from argon2 import PasswordHasher as ArgonPasswordHasher
from argon2.exceptions import VerifyMismatchError

@pytest.fixture
def password_hasher():
    pwd_hasher = PasswordHasher()
    pwd_hasher.crypt_context = CryptContext(schemes=["argon2", ], deprecated="auto")
    yield pwd_hasher


@pytest.fixture
def db_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        with tempfile.NamedTemporaryFile(mode='wb', delete=True) as tmp:
            yield Path(temp_dir) / f"{tmp.name}.db"


@pytest.fixture
def db_connection(db_file):
    test_db_connect = f"sqlite+aiosqlite:///{db_file}"
    yield test_db_connect


@pytest.fixture
@pytest.mark.asyncio
async def user_manager(password_hasher, db_connection):
    um = UserManager(db_connection)
    await um.db_sync()
    yield um


@pytest.fixture
@pytest.mark.asyncio
async def topic_manager(password_hasher, db_connection):
    tm = TopicManager(db_connection)
    await tm.db_sync()
    yield tm


# ######################################
# Tests for the UserAuthDBPlugin


@pytest.mark.asyncio
async def test_create_user(user_manager, db_file, db_connection):
    await user_manager.create_user_auth("myuser", "mypassword")

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        has_user = False
        async with await db_conn.execute("SELECT * FROM user_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] == "myuser"
                assert row['password_hash'] != "mypassword"
                assert '$argon2' in row['password_hash']
                ph = ArgonPasswordHasher()
                ph.verify(row['password_hash'], "mypassword")

                with pytest.raises(VerifyMismatchError):
                    ph.verify(row['password_hash'], "mywrongpassword")

                has_user = True

        assert has_user, "user was not created"


@pytest.mark.asyncio
async def test_list_users(user_manager, db_file, db_connection):
    await user_manager.create_user_auth("myuser", "mypassword")
    await user_manager.create_user_auth("otheruser", "mypassword")
    await user_manager.create_user_auth("anotheruser", "mypassword")

    assert len(list(await user_manager.list_user_auths())) == 3


@pytest.mark.asyncio
async def test_list_empty_users(user_manager, db_file, db_connection):

    assert len(list(await user_manager.list_user_auths())) == 0


@pytest.mark.asyncio
async def test_password_change(user_manager, db_file, db_connection):
    new_user = await user_manager.create_user_auth("myuser", "mypassword")

    await user_manager.update_user_auth_password("myuser", "mynewpassword")
    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        has_user = False
        async with await db_conn.execute("SELECT * FROM user_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['password_hash'] != new_user._password_hash
                ph = ArgonPasswordHasher()
                with pytest.raises(VerifyMismatchError):
                    ph.verify(row['password_hash'], "mypassword")

                has_user = True

    assert has_user, "user was not found"


@pytest.mark.asyncio
async def test_remove_users(user_manager, db_file, db_connection):
    await user_manager.create_user_auth("myuser", "mypassword")
    await user_manager.create_user_auth("otheruser", "mypassword")
    await user_manager.create_user_auth("anotheruser", "mypassword")

    assert len(list(await user_manager.list_user_auths())) == 3

    await user_manager.delete_user_auth("myuser")

    assert len(list(await user_manager.list_user_auths())) == 2

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        test_run = False
        async with await db_conn.execute("SELECT * FROM user_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] in ("otheruser", "anotheruser")
                test_run = True

        assert test_run, "users weren't not found"


@pytest.mark.parametrize("user_pwd,session_pwd,outcome", [
    ("mypassword", "mypassword", True),
    ("mypassword", "myotherpassword", False),
])
@pytest.mark.asyncio
async def test_db_auth(db_connection, user_manager, user_pwd, session_pwd, outcome):

    await user_manager.create_user_auth("myuser", user_pwd)

    broker_context = BrokerContext(broker=Broker())
    broker_context.config = UserAuthDBPlugin.Config(
        connection=db_connection
    )
    db_auth_plugin = UserAuthDBPlugin(context=broker_context)

    s = Session()
    s.username = "myuser"
    s.password = session_pwd

    assert await db_auth_plugin.authenticate(session=s) == outcome


@pytest.mark.asyncio
async def test_client_authentication(user_manager, db_connection):

    user = await user_manager.create_user_auth("myuser", "mypassword")
    assert user is not None

    broker_cfg = {
        'listeners': { 'default': {'type': 'tcp', 'bind': '127.0.0.1:1883'}},
        'plugins': {
            'amqtt.contrib.auth_db.UserAuthDBPlugin': {
                'connection': db_connection,
            }
        }
    }

    broker = Broker(config=broker_cfg)

    await broker.start()
    await asyncio.sleep(0.1)

    client = MQTTClient(client_id='myclientid', config={'auto_reconnect': False})
    await client.connect("mqtt://myuser:mypassword@127.0.0.1:1883")

    await client.subscribe([
        ("my/topic", QOS_1)
    ])
    await asyncio.sleep(0.1)
    await client.publish("my/topic", b"test")
    await asyncio.sleep(0.1)
    message = await client.deliver_message(timeout_duration=1)
    assert message.topic == "my/topic"
    await asyncio.sleep(0.1)
    await client.disconnect()
    await asyncio.sleep(0.1)
    await broker.shutdown()


@pytest.mark.parametrize("client_pwd", [
    ("mywrongpassword", ),
    ("", ),
])
@pytest.mark.asyncio
async def test_client_blocked(user_manager, db_connection, client_pwd):

    user = await user_manager.create_user_auth("myuser", "mypassword")
    assert user is not None

    broker_cfg = {
        'listeners': { 'default': {'type': 'tcp', 'bind': '127.0.0.1:1883'}},
        'plugins': {
            'amqtt.contrib.auth_db.UserAuthDBPlugin': {
                'connection': db_connection,
            }
        }
    }

    broker = Broker(config=broker_cfg)

    await broker.start()
    await asyncio.sleep(0.1)

    client = MQTTClient(client_id='myclientid', config={'auto_reconnect': False})
    with pytest.raises(ConnectError):
        await client.connect(f"mqtt://myuser:{client_pwd}@127.0.0.1:1883")

    await broker.shutdown()
    await asyncio.sleep(0.1)


# ######################################
# Tests for the TopicAuthDBPlugin

def test_allowed_topic_match():

    at = AllowedTopic(topic="my/topic")
    assert "my/topic" in at

    at2 = AllowedTopic(topic="my/other/topic")
    assert "my/other/topic" in at2
    assert "my/another/topic" not in at2

    at3 = AllowedTopic(topic="my/#")
    assert "my/other" in at3
    assert "my/other/topic" in at3
    assert "other/topic" not in at3

    at4 = AllowedTopic(topic="my/other/#")
    assert "my/other" in at4
    assert "my/other/topic" in at4
    assert "other/topic" not in at4
    assert "/my/other/topic" not in at4

    at5 = AllowedTopic(topic="my/+/topic")
    assert "my/other/topic" in at5
    assert "my/another/topic" in at5
    assert "my/other/another/topic" not in at5

    at6 = AllowedTopic(topic="my/other/topic")
    assert at6 == at2
    assert at2 == at6
    assert at6 != at
    assert at6 not in at5


def test_allowed_topic_list_match():
    at1 = AllowedTopic(topic="one/topic")
    at2 = AllowedTopic(topic="one/other/topic")
    at3 = AllowedTopic(topic="two/+/topic")
    at4 = AllowedTopic(topic="three/topic/#")

    at_list = [at1, at2, at3, at4]

    assert "one/topic" in at_list
    assert "two/other/topic" in at_list
    assert "two/another/topic" in at_list
    assert "three/topic" in at_list
    assert "three/topic/other" in at_list


def test_remove_topic_list():
    at1 = AllowedTopic(topic="my/topic")
    at2 = AllowedTopic(topic="my/other/topic")
    at3 = AllowedTopic(topic="my/another/topic")

    at_list = [at1, at2, at3]

    at4 = AllowedTopic(topic="my/topic")
    at5 = AllowedTopic(topic="my/not/topic")
    at_list.remove(at4)

    with pytest.raises(ValueError):
        at_list.remove(at5)

@pytest.mark.asyncio
async def test_add_topic_to_client(db_file, user_manager, topic_manager, db_connection):
    client_id = "myuser"

    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None
    topic_list = await topic_manager.add_allowed_topic(client_id, "my/topic", Action.PUBLISH)
    assert len(topic_list) > 0

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        user_found = False
        async with await db_conn.execute("SELECT * FROM topic_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] == client_id
                assert "my/topic" in row['publish_acl']
                user_found = True

        assert user_found


@pytest.mark.asyncio
async def test_invalid_dollar_topic_for_publish(db_file, user_manager, topic_manager, db_connection):
    client_id = "myuser"

    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None
    with pytest.raises(MQTTError):
        await topic_manager.add_allowed_topic(client_id, "$MY/topic", Action.PUBLISH)

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        auth_topic_found = False
        async with await db_conn.execute("SELECT * FROM topic_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] == client_id
                assert "$MY/topic" not in row['publish_acl']
                auth_topic_found = True

        assert auth_topic_found


@pytest.mark.asyncio
async def test_remove_topic_from_client(db_file, user_manager, topic_manager, db_connection):
    client_id = "myuser"
    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None
    user = await user_manager.create_user_auth(client_id, "mypassword")
    assert user is not None

    await topic_manager.add_allowed_topic(client_id, "my/topic", Action.PUBLISH)
    await topic_manager.add_allowed_topic(client_id, "my/other/topic", Action.PUBLISH)
    topic_list = await topic_manager.add_allowed_topic(client_id, "my/another/topic", Action.PUBLISH)
    assert len(topic_list) == 3

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        user_found = False
        async with await db_conn.execute("SELECT * FROM topic_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] == client_id
                assert "my/topic" in row['publish_acl']
                assert "my/other/topic" in row['publish_acl']
                assert "my/another/topic" in row['publish_acl']
                user_found = True
        assert user_found

    topic_list = await topic_manager.remove_allowed_topic(client_id, "my/other/topic", Action.PUBLISH)
    assert len(topic_list) == 2

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        user_found = False
        async with await db_conn.execute("SELECT * FROM topic_auth") as cursor:
            for row in await cursor.fetchall():
                assert row['username'] == client_id
                assert "my/topic" in row['publish_acl']
                assert "my/other/topic" not in row['publish_acl']
                assert "my/another/topic" in row['publish_acl']
                user_found = True
        assert user_found


@pytest.mark.asyncio
async def test_remove_missing_topic(db_file, user_manager, topic_manager, db_connection):
    client_id = "myuser"
    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None
    user = await user_manager.create_user_auth(client_id, "mypassword")
    assert user is not None

    await topic_manager.add_allowed_topic(client_id, "my/topic", Action.PUBLISH)
    await topic_manager.add_allowed_topic(client_id, "my/other/topic", Action.PUBLISH)
    topic_list = await topic_manager.add_allowed_topic(client_id, "my/another/topic", Action.PUBLISH)
    assert len(topic_list) == 3

    with pytest.raises(MQTTError):
        await topic_manager.remove_allowed_topic(client_id, "my/not/topic", Action.PUBLISH)


@pytest.mark.asyncio
async def test_remove_topic_wrong_action(db_file, user_manager, topic_manager, db_connection):
    client_id = "myuser"
    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None
    user = await user_manager.create_user_auth(client_id, "mypassword")
    assert user is not None

    await topic_manager.add_allowed_topic(client_id, "my/topic", Action.PUBLISH)
    await topic_manager.add_allowed_topic(client_id, "my/other/topic", Action.PUBLISH)
    topic_list = await topic_manager.add_allowed_topic(client_id, "my/another/topic", Action.PUBLISH)
    assert len(topic_list) == 3

    with pytest.raises(MQTTError):
        await topic_manager.remove_allowed_topic(client_id, "my/other/topic", Action.SUBSCRIBE)


@pytest.mark.parametrize("acl_topic,msg_topic,outcome", [
    ("my/topic", "my/topic", True),
    ("my/topic", "my/other/topic", False),
    ("my/#", "my/other/topic", True),
    ("my/#", "my/another/topic", True),
])
@pytest.mark.asyncio
async def test_topic_publish_filter_plugin(db_file, topic_manager, db_connection, acl_topic, msg_topic, outcome):
    client_id = "myuser"

    user = await topic_manager.create_topic_auth(client_id)
    assert user is not None

    await topic_manager.add_allowed_topic(client_id, acl_topic, Action.PUBLISH)

    broker_context = BrokerContext(broker=Broker())
    broker_context.config = TopicAuthDBPlugin.Config(
        connection=db_connection
    )
    db_auth_plugin = TopicAuthDBPlugin(context=broker_context)

    s = Session()
    s.username = client_id
    assert await db_auth_plugin.topic_filtering(session=s, topic=msg_topic, action=Action.PUBLISH) == outcome,\
        f"topic filter responded incorrectly: {not outcome} vs {outcome}."



@pytest.mark.asyncio
async def test_topic_subscribe(db_file, topic_manager, db_connection):

    broker_cfg = {
        'listeners': { 'default': {'type': 'tcp', 'bind': '127.0.0.1:1883'}},
        'plugins': {
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {},
            'amqtt.contrib.auth_db.TopicAuthDBPlugin': {
                'connection': db_connection
            },
            'amqtt.plugins.sys.broker.BrokerSysPlugin': {
                'sys_interval' : 2
            }
        }
    }
    client_id = "myuser"

    topic_auth = await topic_manager.create_topic_auth(client_id)
    assert topic_auth is not None

    sub_topic_list = await topic_manager.add_allowed_topic(client_id, '$SYS/#', Action.SUBSCRIBE)
    rcv_topic_list = await topic_manager.add_allowed_topic(client_id, '$SYS/#', Action.RECEIVE)

    assert len(sub_topic_list) > 0
    assert len(rcv_topic_list) > 0

    broker = Broker(config=broker_cfg)

    await broker.start()
    await asyncio.sleep(0.1)

    client = MQTTClient(client_id='myclientid', config={'auto_reconnect': False})
    await client.connect("mqtt://myuser:mypassword@127.0.0.1:1883")

    ret = await client.subscribe([
        ('$SYS/broker/clients/connected', QOS_0)
    ])

    assert ret == [0x0,]

    await asyncio.sleep(0.1)

    message_received = False
    try:
        message = await client.deliver_message(timeout_duration=4)
        assert message.topic == '$SYS/broker/clients/connected'
        message_received = True
    except asyncio.TimeoutError:
        pass

    assert message_received, "Did not receive a $SYS message"
    await asyncio.sleep(0.1)
    await client.disconnect()
    await asyncio.sleep(0.1)
    await broker.shutdown()



import asyncio
import logging
from importlib.metadata import EntryPoint
from unittest.mock import patch

import pytest

from amqtt.broker import Broker
from amqtt.client import MQTTClient
from amqtt.errors import ClientError, ConnectError, MQTTError
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.ERROR, format=formatter)
log = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_connect_tcp(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://localhost:1883/")
    assert client.session is not None
    await client.disconnect()

@pytest.mark.asyncio
async def test_connect_tcp_secure(rsa_keys, broker_fixture):
    certfile, _ = rsa_keys
    client = MQTTClient(config={"check_hostname": False, "auto_reconnect": False})

    # since we're using a self-signed certificate, need to provide the server's certificate to verify authenticity
    await client.connect("mqtts://localhost:1884/", cafile=certfile)
    assert client.session is not None
    await client.disconnect()


@pytest.mark.asyncio
async def test_connect_tcp_failure():
    config = {"auto_reconnect": False}
    client = MQTTClient(config=config)
    with pytest.raises(ConnectError):
        await client.connect("mqtt://127.0.0.1/")


@pytest.mark.asyncio
async def test_connect_ws(broker_fixture):
    client = MQTTClient()
    await client.connect("ws://127.0.0.1:8080/")
    assert client.session is not None
    await client.disconnect()


@pytest.mark.asyncio
async def test_reconnect_ws_retain_username_password(broker_fixture):
    client = MQTTClient()
    await client.connect("ws://fred:password@127.0.0.1:8080/")
    assert client.session is not None
    assert client.session.username is not None
    assert client.session.password is not None
    await client.disconnect()
    await client.reconnect()

    assert client.session.username is not None
    assert client.session.password is not None


@pytest.mark.asyncio
async def test_connect_ws_secure(rsa_keys, broker_fixture):
    certfile, _ = rsa_keys
    client = MQTTClient(config={"auto_reconnect": False})
    # since we're using a self-signed certificate, need to provide the server's certificate to verify authenticity
    await client.connect("wss://localhost:8081/", cafile=certfile)
    assert client.session is not None
    await client.disconnect()


@pytest.mark.asyncio
async def test_connect_username_without_password(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://alice@127.0.0.1/")
    assert client.session is not None
    await client.disconnect()


@pytest.mark.asyncio
async def test_ping(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1/")
    assert client.session is not None
    await client.ping()
    await client.disconnect()


@pytest.mark.asyncio
async def test_subscribe(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1/")
    assert client.session is not None
    ret = await client.subscribe(
        [
            ("$SYS/broker/uptime", QOS_0),
            ("$SYS/broker/uptime", QOS_1),
            ("$SYS/broker/uptime", QOS_2),
        ],
    )
    assert ret[0] == QOS_0
    assert ret[1] == QOS_1
    assert ret[2] == QOS_2
    await client.disconnect()


@pytest.mark.asyncio
async def test_unsubscribe(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1/")
    assert client.session is not None
    ret = await client.subscribe(
        [
            ("$SYS/broker/uptime", QOS_0),
        ],
    )
    assert ret[0] == QOS_0
    await client.unsubscribe(["$SYS/broker/uptime"])
    await client.disconnect()


@pytest.mark.asyncio
async def test_deliver(broker_fixture):
    data = b"data"
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1/")
    assert client.session is not None
    ret = await client.subscribe(
        [
            ("test_topic", QOS_0),
        ],
    )
    assert ret[0] == QOS_0
    client_pub = MQTTClient()
    await client_pub.connect("mqtt://127.0.0.1/")
    await client_pub.publish("test_topic", data, QOS_0)
    await client_pub.disconnect()
    message = await client.deliver_message()
    assert message is not None
    assert message.publish_packet is not None
    assert message.data == data
    await client.unsubscribe(["$SYS/broker/uptime"])
    await client.disconnect()


@pytest.mark.asyncio
async def test_deliver_timeout(broker_fixture):
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1/")
    assert client.session is not None
    ret = await client.subscribe(
        [
            ("test_topic", QOS_0),
        ],
    )
    assert ret[0] == QOS_0
    with pytest.raises(asyncio.TimeoutError):
        await client.deliver_message(timeout_duration=2)
    await client.unsubscribe(["$SYS/broker/uptime"])
    await client.disconnect()


@pytest.mark.asyncio
async def test_cancel_publish_qos1(broker_fixture):
    """Tests that timeouts on published messages will clean up in-flight messages."""
    data = b"data"
    client_pub = MQTTClient()
    await client_pub.connect("mqtt://127.0.0.1/")

    assert client_pub.session is not None
    assert client_pub._handler is not None

    assert client_pub.session.inflight_out_count == 0
    fut = asyncio.create_task(client_pub.publish("test_topic", data, QOS_1))
    assert len(client_pub._handler._puback_waiters) == 0
    while len(client_pub._handler._puback_waiters) == 0 and not fut.done():
        await asyncio.sleep(0)
    assert len(client_pub._handler._puback_waiters) == 1
    assert client_pub.session.inflight_out_count == 1
    fut.cancel()
    await asyncio.wait([fut])
    assert len(client_pub._handler._puback_waiters) == 0
    assert client_pub.session.inflight_out_count == 0

    await asyncio.sleep(0.1)
    await client_pub.disconnect()


@pytest.mark.asyncio
async def test_cancel_publish_qos2_pubrec(broker_fixture):
    """Tests that timeouts on published messages will clean up in-flight messages."""
    data = b"data"
    client_pub = MQTTClient()
    await client_pub.connect("mqtt://127.0.0.1/")

    assert client_pub.session is not None
    assert client_pub._handler is not None

    assert client_pub.session.inflight_out_count == 0
    fut = asyncio.create_task(client_pub.publish("test_topic", data, QOS_2))
    assert len(client_pub._handler._pubrec_waiters) == 0
    while len(client_pub._handler._pubrec_waiters) == 0 or fut.done() or fut.cancelled():
        await asyncio.sleep(0)
    assert len(client_pub._handler._pubrec_waiters) == 1
    assert client_pub.session.inflight_out_count == 1
    fut.cancel()
    await asyncio.sleep(1)
    await asyncio.wait([fut])
    assert len(client_pub._handler._pubrec_waiters) == 0
    assert client_pub.session.inflight_out_count == 0

    await asyncio.sleep(0.1)
    await client_pub.disconnect()


@pytest.mark.asyncio
async def test_cancel_publish_qos2_pubcomp(broker_fixture):
    """Tests that timeouts on published messages will clean up in-flight messages."""
    data = b"data"
    client_pub = MQTTClient()
    await client_pub.connect("mqtt://127.0.0.1/")

    assert client_pub.session is not None
    assert client_pub._handler is not None

    assert client_pub.session.inflight_out_count == 0
    fut = asyncio.create_task(client_pub.publish("test_topic", data, QOS_2))
    assert len(client_pub._handler._pubcomp_waiters) == 0
    while len(client_pub._handler._pubcomp_waiters) == 0 and not fut.done():
        await asyncio.sleep(0)
    assert len(client_pub._handler._pubcomp_waiters) == 1
    fut.cancel()
    await asyncio.wait([fut])
    assert len(client_pub._handler._pubcomp_waiters) == 0
    assert client_pub.session.inflight_out_count == 0

    await asyncio.sleep(0.1)
    await client_pub.disconnect()


@pytest.fixture
def client_config():
    return {
        "default_retain": False,
        "topics": {
                "test": {
                    "qos": 0
                },
                "some_topic": {
                    "retain": True,
                    "qos": 2
                }
        },
        "keep_alive": 10,
        "connection": {
            "uri": "mqtt://localhost:1884"
        },
        "reconnect_max_interval": 5,
        "will": {
            "topic": "test/will/topic",
            "retain": True,
            "message": "client ABC has disconnected",
            "qos": 1
        },
        "ping_delay": 1,
        "default_qos": 0,
        "auto_reconnect": True,
        "reconnect_retries": 10
    }


@pytest.mark.asyncio
async def test_client_will_with_clean_disconnect(broker_fixture):
    config = {
        "will": {
            "topic": "test/will/topic",
            "retain": False,
            "message": "client ABC has disconnected",
            "qos": 1
        },
    }

    client1 = MQTTClient(client_id="client1", config=config)
    await client1.connect("mqtt://localhost:1883")

    client2 = MQTTClient(client_id="client2")
    await client2.connect("mqtt://localhost:1883")
    await client2.subscribe(
        [
            ("test/will/topic", QOS_0),
        ]
    )

    await client1.disconnect()
    await asyncio.sleep(1)

    with pytest.raises(asyncio.TimeoutError):
        message = await client2.deliver_message(timeout_duration=2)
        # if we do get a message, make sure it's not a will message
        assert message.topic != "test/will/topic"

    await client2.disconnect()


@pytest.mark.asyncio
async def test_client_will_with_abrupt_disconnect(broker_fixture):
    config = {
        "will": {
            "topic": "test/will/topic",
            "retain": False,
            "message": "client ABC has disconnected",
            "qos": 1
        },
    }

    client1 = MQTTClient(client_id="client1", config=config)
    await client1.connect("mqtt://localhost:1883")

    client2 = MQTTClient(client_id="client2")
    await client2.connect("mqtt://localhost:1883")
    await client2.subscribe(
        [
            ("test/will/topic", QOS_0),
        ]
    )

    # instead of client.disconnect, call the necessary closing but without sending the disconnect packet
    await client1.cancel_tasks()
    if client1._disconnect_task and not client1._disconnect_task.done():
        client1._disconnect_task.cancel()
    client1._connected_state.clear()
    await client1._handler.stop()
    client1.session.transitions.disconnect()

    await asyncio.sleep(1)


    message = await client2.deliver_message(timeout_duration=1)
    # make sure we receive the will message
    assert message.topic == "test/will/topic"
    assert message.data == b'client ABC has disconnected'

    await client2.disconnect()


@pytest.mark.asyncio
async def test_client_retained_will_with_abrupt_disconnect(broker_fixture):

    # verifying client functionality of retained will topic/message

    config = {
        "will": {
            "topic": "test/will/topic",
            "retain": True,
            "message": "client ABC has disconnected",
            "qos": 1
        },
    }

    # first client, connect with retained will message
    client1 = MQTTClient(client_id="client1", config=config)
    await client1.connect('mqtt://localhost:1883')

    client2 = MQTTClient(client_id="client2")
    await client2.connect('mqtt://localhost:1883')
    await  client2.subscribe([
        ("test/will/topic", QOS_0)
        ])


    # let's abruptly disconnect client1
    await client1.cancel_tasks()
    if client1._disconnect_task and not client1._disconnect_task.done():
        client1._disconnect_task.cancel()
    client1._connected_state.clear()
    await client1._handler.stop()
    client1.session.transitions.disconnect()

    await asyncio.sleep(0.5)

    # make sure the client which is still connected that we get the 'will' message
    message = await client2.deliver_message(timeout_duration=1)
    assert message.topic == 'test/will/topic'
    assert message.data == b'client ABC has disconnected'
    await client2.disconnect()

    # make sure a client which is connected after client1 disconnected still receives the 'will' message from
    client3 = MQTTClient(client_id="client3")
    await client3.connect('mqtt://localhost:1883')
    await client3.subscribe([
        ("test/will/topic", QOS_0)
    ])
    message3 = await client3.deliver_message(timeout_duration=1)
    assert message3.topic == 'test/will/topic'
    assert message3.data == b'client ABC has disconnected'
    await client3.disconnect()


@pytest.mark.asyncio
async def test_client_abruptly_disconnecting_with_empty_will_message(broker_fixture):

    config = {
        "will": {
            "topic": "test/will/topic",
            "retain": True,
            "message": "",
            "qos": 1
        },
    }
    client1 = MQTTClient(client_id="client1", config=config)
    await client1.connect('mqtt://localhost:1883')

    client2 = MQTTClient(client_id="client2")
    await client2.connect('mqtt://localhost:1883')
    await client2.subscribe([
        ("test/will/topic", QOS_0)
    ])

    # let's abruptly disconnect client1
    await client1.cancel_tasks()
    if client1._disconnect_task and not client1._disconnect_task.done():
        client1._disconnect_task.cancel()
    client1._connected_state.clear()
    await client1._handler.stop()
    client1.session.transitions.disconnect()

    await asyncio.sleep(0.5)

    message = await client2.deliver_message(timeout_duration=1)
    assert message.topic == 'test/will/topic'
    assert message.data == b''

    await client2.disconnect()


async def test_connect_broken_uri():
    config = {"auto_reconnect": False}
    client = MQTTClient(config=config)
    with pytest.raises(ClientError):
        await client.connect('"mqtt://someplace')


@pytest.mark.asyncio
async def test_connect_incorrect_scheme():
    config = {"auto_reconnect": False}
    client = MQTTClient(config=config)
    with pytest.raises(ClientError):
        await client.connect('"mq://someplace')


@pytest.mark.asyncio
@pytest.mark.timeout(3)
async def test_connect_timeout():
    config = {"auto_reconnect": False, "connection_timeout": 2}
    client = MQTTClient(config=config)
    with pytest.raises(ClientError):
        await client.connect("mqtt://localhost:8888")


async def test_client_no_auth():

    class MockEntryPoints:

        def select(self, group) -> list[EntryPoint]:
            match group:
                case 'tests.mock_plugins':
                    return [
                            EntryPoint(name='auth_plugin', group='tests.mock_plugins', value='tests.plugins.mocks:TestNoAuthPlugin'),
                        ]
                case _:
                    return list()


    with patch("amqtt.plugins.manager.entry_points", side_effect=MockEntryPoints) as mocked_mqtt_publish:

        config = {
            "listeners": {
                "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
            },
            'sys_interval': 1,
            'auth': {
                'plugins': ['auth_plugin', ]
            }
        }

        client = MQTTClient(client_id="client1", config={'auto_reconnect': False})

        with pytest.warns(DeprecationWarning):

            broker = Broker(plugin_namespace='tests.mock_plugins', config=config)
            await broker.start()

            with pytest.raises(ConnectError):
                await client.connect("mqtt://127.0.0.1:1883/")

            await broker.shutdown()


@pytest.mark.asyncio
async def test_publish_to_incorrect_wildcard(broker_fixture):
    client = MQTTClient(config={'auto_reconnect': False})
    await client.connect("mqtt://127.0.0.1/")

    with pytest.raises(MQTTError):
        await client.publish("my/+/topic", b'plus-sign wildcard topic invalid publish')
    with pytest.raises(MQTTError):
        await client.publish("topic/#", b'hash wildcard topic invalid publish')

    await client.publish("topic/*", b'asterisk topic normal publish')
    await client.disconnect()



import asyncio
import ldap
import time
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from amqtt.broker import BrokerContext, Broker
from amqtt.client import MQTTClient
from amqtt.contexts import BrokerConfig, ListenerConfig, ClientConfig, Action
from amqtt.contrib.ldap import UserAuthLdapPlugin, TopicAuthLdapPlugin
from amqtt.errors import ConnectError
from amqtt.session import Session


# Pin the project name to avoid creating multiple stacks
@pytest.fixture(scope="session")
def docker_compose_project_name() -> str:
    return "openldap"

# Stop the stack before starting a new one
@pytest.fixture(scope="session")
def docker_setup():
    return ["down -v", "up --build -d"]

@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return Path(pytestconfig.rootdir) / "tests/fixtures/ldap" / "docker-compose.yml"

@pytest.fixture(scope="session")
def ldap_service_docker(docker_ip, docker_services):
    """Ensure that HTTP service is up and responsive."""

    # `port_for` takes a container port and returns the corresponding host port
    port = docker_services.port_for("openldap", 389)
    url = "ldap://{}:{}".format(docker_ip, port)
    time.sleep(2)
    return url

@pytest.fixture(scope="session")
def ldap_service_mock():
    """Ensure that HTTP service is up and responsive."""

    # `port_for` takes a container port and returns the corresponding host port
    url = "ldap://localhost:0"
    mock_ldap_object = MagicMock()

    def mock_simple_s(*args, **kwargs):
        return [ ('dn', {'uid':'jdoe', 'publishACL':[b'my/topic/one', b'my/topic/two']}), ]

    def mock_simple_bind_s(*args):
        p = args[1]
        if p in ("adminpassword", "johndoepassword"):
            return
        raise ldap.INVALID_CREDENTIALS

    mock_ldap_object.search_s.side_effect = mock_simple_s
    mock_ldap_object.simple_bind_s.side_effect = mock_simple_bind_s
    with patch("ldap.initialize", return_value=mock_ldap_object):
        yield url

@pytest.fixture(scope="session")
def ldap_service(request):
    mode = request.config.getoption("--mock-docker")
    if mode:
        return request.getfixturevalue("ldap_service_mock")
    return request.getfixturevalue("ldap_service_docker")


@pytest.mark.asyncio
async def test_ldap_user_plugin(ldap_service):
    ctx = BrokerContext(Broker())
    ctx.config = UserAuthLdapPlugin.Config(
        server=ldap_service,
        base_dn="dc=amqtt,dc=io",
        user_attribute="uid",
        bind_dn="cn=admin,dc=amqtt,dc=io",
        bind_password="adminpassword",
    )
    ldap_plugin = UserAuthLdapPlugin(context=ctx)

    s = Session()
    s.username = "jdoe"
    s.password = "johndoepassword"

    assert await ldap_plugin.authenticate(session=s), "could not authenticate user"

@pytest.mark.asyncio
async def test_ldap_user(ldap_service):

    cfg = BrokerConfig(
        listeners={ 'default' : ListenerConfig() },
        plugins={
            'amqtt.contrib.ldap.UserAuthLdapPlugin': {
                'server': ldap_service,
                'base_dn': 'dc=amqtt,dc=io',
                'user_attribute': 'uid',
                'bind_dn': 'cn=admin,dc=amqtt,dc=io',
                'bind_password': 'adminpassword',
            },
        }
    )

    broker = Broker(config=cfg)
    await broker.start()

    await asyncio.sleep(0.1)

    client = MQTTClient(config=ClientConfig(auto_reconnect=False))
    await client.connect('mqtt://jdoe:johndoepassword@127.0.0.1:1883')
    await asyncio.sleep(0.1)
    await client.publish('my/topic', b'my message')
    await asyncio.sleep(0.1)
    await client.disconnect()
    await broker.shutdown()


@pytest.mark.asyncio
async def test_ldap_user_invalid_creds(ldap_service):

    cfg = BrokerConfig(
        listeners={ 'default' : ListenerConfig() },
        plugins={
            'amqtt.contrib.ldap.UserAuthLdapPlugin': {
                'server': ldap_service,
                'base_dn': 'dc=amqtt,dc=io',
                'user_attribute': 'uid',
                'bind_dn': 'cn=admin,dc=amqtt,dc=io',
                'bind_password': 'adminpassword',
            },
        }
    )

    broker = Broker(config=cfg)
    await broker.start()

    await asyncio.sleep(0.1)

    client = MQTTClient(config=ClientConfig(auto_reconnect=False))
    with pytest.raises(ConnectError):
        await client.connect('mqtt://jdoe:wrongpassword@127.0.0.1:1883')

    await broker.shutdown()


@pytest.mark.asyncio
async def test_ldap_topic_plugin(ldap_service):
    ctx = BrokerContext(Broker())
    ctx.config = TopicAuthLdapPlugin.Config(
        server=ldap_service,
        base_dn="dc=amqtt,dc=io",
        user_attribute="uid",
        bind_dn="cn=admin,dc=amqtt,dc=io",
        bind_password="adminpassword",
        publish_attribute="publishACL",
        subscribe_attribute="subscribeACL",
        receive_attribute="receiveACL"
    )
    ldap_plugin = TopicAuthLdapPlugin(context=ctx)

    s = Session()
    s.username = "jdoe"
    s.password = "wrongpassword"

    assert await ldap_plugin.topic_filtering(session=s, topic='my/topic/one', action=Action.PUBLISH), "access not granted"



import asyncio
import contextlib
from dataclasses import dataclass
import json
import logging
import os
from pathlib import Path
import socket
import sys
from typing import Any

import typer

from amqtt import __version__ as amqtt_version
from amqtt.client import MQTTClient
from amqtt.errors import ClientError, ConnectError, MQTTError
from amqtt.mqtt.constants import QOS_0
from amqtt.utils import read_yaml_config

logger = logging.getLogger(__name__)


def _gen_client_id() -> str:
    pid = os.getpid()
    hostname = socket.gethostname()
    return f"amqtt_sub/{pid}-{hostname}"


def _get_extra_headers(extra_headers_json: str | None = None) -> dict[str, Any]:
    try:
        extra_headers: dict[str, Any] = json.loads(extra_headers_json or "{}")
    except (json.JSONDecodeError, TypeError):
        return {}
    return extra_headers


@dataclass
class CAInfo:
    ca_file: str | None = None
    ca_path: str | None = None
    ca_data: str | None = None


async def do_sub(client: MQTTClient,
                 url: str,
                 topics: list[str],
                 ca_info: CAInfo,
                 max_count: int | None = None,
                 clean_session: bool = False,
                 extra_headers_json: str | None = None,
                 qos: int | None = None,
                 ) -> None:
    """Perform the subscription."""
    try:
        logger.info(f"{client.client_id} Connecting to broker")

        await client.connect(
            uri=url,
            cleansession=clean_session,
            cafile=ca_info.ca_file,
            capath=ca_info.ca_path,
            cadata=ca_info.ca_data,
            additional_headers=_get_extra_headers(extra_headers_json),
        )

        filters = [(topic, qos) for topic in topics]
        await client.subscribe(filters)

        count = 0
        while True:
            if max_count and count >= max_count:
                break
            try:
                message = await client.deliver_message()
                if message and message.publish_packet and message.publish_packet.data:
                    count += 1
                    sys.stdout.buffer.write(message.publish_packet.data)
                    sys.stdout.write("\n")
            except MQTTError:
                logger.debug("Error reading packet")

        await client.disconnect()
        logger.info(f"{client.client_id} Disconnected from broker")
    except KeyboardInterrupt:
        await client.disconnect()
        logger.info(f"{client.client_id} Disconnected from broker")
    except ConnectError as exc:
        logger.fatal(f"Connection to '{url}' failed: {exc!r}")
        raise ConnectError from exc
    except asyncio.CancelledError as exc:
        logger.fatal("Publish canceled due to previous error")
        raise asyncio.CancelledError from exc


app = typer.Typer(add_completion=False, rich_markup_mode=None)


def main() -> None:
    """Entry point for the amqtt subscriber."""
    app()


def _version(v: bool) -> None:
    if v:
        typer.echo(f"{amqtt_version}")
        raise typer.Exit(code=0)


@app.command()
def subscribe_main(  # pylint: disable=R0914,R0917
    url: str = typer.Option(None, help="Broker connection URL, *must conform to MQTT or URI scheme: `[mqtt(s)|ws(s)]://<username:password>@HOST:port`*", show_default=False),
    config_file: str | None = typer.Option(None, "-c", help="Client configuration file"),
    client_id: str | None = typer.Option(None, "-i", "--client-id", help="client identification for mqtt connection. *default: process id and the hostname of the client*"), max_count: int | None = typer.Option(None, "-n", help="Number of messages to read before ending *default: read indefinitely*"),
    qos: int = typer.Option(0, "--qos", "-q", help="Quality of service (0, 1, or 2)"),
    topics: list[str] = typer.Option(..., "-t", help="Topic filter to subscribe, can be used multiple times."),  # noqa: B008
    keep_alive: int | None = typer.Option(None, "-k", help="Keep alive timeout in seconds"),
    clean_session: bool = typer.Option(False, "--clean-session", help="Clean session on connect. *default: False*"),
    ca_file: str | None = typer.Option(None, "--ca-file", help="Define the path to a file containing PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    ca_path: str | None = typer.Option(None, "--ca-path", help="Define the path to a directory containing PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    ca_data: str | None = typer.Option(None, "--ca-data", help="Set the PEM encoded CA certificates that are trusted. Used to enable SSL communication."),
    will_topic: str | None = typer.Option(None, "--will-topic", help="The topic on which to send a Will, in the event that the client disconnects unexpectedly."),
    will_message: str | None = typer.Option(None, "--will-message", help="Specify a message that will be stored by the broker and sent out if this client disconnects unexpectedly. *required if `--will-topic` is specified*."),
    will_qos: int = typer.Option(0, "--will-qos", help="The QoS to use for the Will. *default: 0, only valid if `--will-topic` is specified*"),
    will_retain: bool = typer.Option(False, "--will-retain", help="If the client disconnects unexpectedly the message sent out will be treated as a retained message. *only valid, if `--will-topic` is specified*"),
    extra_headers_json: str | None = typer.Option(None, "--extra-headers", help="Specify a JSON object string with key-value pairs representing additional headers that are transmitted on the initial connection. *websocket connections only*."),
    debug: bool = typer.Option(False, "-d", help="Enable debug messages"),
    version: bool = typer.Option(False, "--version", callback=_version, is_eager=True, help="Show version and exit"),  # noqa : ARG001
) -> None:
    """Command line MQTT client to subscribe to one or more topics and display any messages received."""
    if bool(will_message) != bool(will_topic):
        typer.echo("❌ must specify both 'will_message' and 'will_topic' ")
        raise typer.Exit(code=1)

    if will_retain and not (will_message and will_topic):
        typer.echo("❌ 'will-retain' only valid if 'will_message' and 'will_topic' are specified.", err=True)
        raise typer.Exit(code=1)

    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level, format=formatter)

    if config_file:
        config = read_yaml_config(config_file)
    else:
        default_config_path = Path(__file__).parent / "default_client.yaml"
        logger.debug(f"Using default configuration from {default_config_path}")
        config = read_yaml_config(default_config_path)

    if not client_id:
        client_id = _gen_client_id()

    if not isinstance(config, dict):
        logger.debug("Failed to correctly initialize config")
        return

    if keep_alive:
        config["keep_alive"] = keep_alive

    if will_topic and will_message:
        config["will"] = {
            "topic": will_topic,
            "message": will_message,
            "qos": int(will_qos),
            "retain": will_retain,
        }

    client = MQTTClient(client_id=client_id, config=config)
    ca_info = CAInfo(
        ca_file=ca_file,
        ca_path=ca_path,
        ca_data=ca_data,
    )
    with contextlib.suppress(KeyboardInterrupt):
        try:
            asyncio.run(do_sub(client,
                                           url=url,
                                           topics=topics,
                                           ca_info=ca_info,
                                           extra_headers_json=extra_headers_json,
                                           qos=qos or QOS_0,
                                           max_count=max_count,
                                           clean_session=clean_session,
                                           ))

        except (ClientError, ConnectError) as exc:
            typer.echo("❌ Connection failed", err=True)
            raise typer.Exit(code=1) from exc


if __name__ == "__main__":
    main()



import asyncio
from asyncio import CancelledError
import logging

from amqtt.client import MQTTClient

"""
This sample shows how to run an idle client
"""

logger = logging.getLogger(__name__)

config = {
    "keep_alive": 5,
    "ping_delay": 1,
}

async def main() -> None:
    client = MQTTClient(config=config)

    try:
        await client.connect("mqtt://localhost:1883/")
        logger.info("client connected")
        await asyncio.sleep(15)
    except CancelledError:
        pass

    await client.disconnect()


def __main__():

    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    asyncio.run(main())


if __name__ == "__main__":
    __main__()



import asyncio
import unittest

from amqtt.adapters import BufferReader
from amqtt.mqtt.constants import QOS_1, QOS_2
from amqtt.mqtt.packet import PacketIdVariableHeader
from amqtt.mqtt.subscribe import SubscribePacket, SubscribePayload


class SubscribePacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b"\x80\x0e\x00\x0a\x00\x03a/b\x01\x00\x03c/d\x02"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(SubscribePacket.from_stream(stream))
        (topic, qos) = message.payload.topics[0]
        assert topic == "a/b"
        assert qos == QOS_1
        (topic, qos) = message.payload.topics[1]
        assert topic == "c/d"
        assert qos == QOS_2

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = SubscribePayload([("a/b", QOS_1), ("c/d", QOS_2)])
        publish = SubscribePacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        assert out == b"\x82\x0e\x00\n\x00\x03a/b\x01\x00\x03c/d\x02"



import pytest
from amqtt.errors import AMQTTError
from amqtt.mqtt.connack import ConnackPacket
from amqtt.mqtt.packet import MQTTFixedHeader, PUBLISH



def test_incorrect_fixed_header():
    header = MQTTFixedHeader(PUBLISH, 0x00)
    with pytest.raises(AMQTTError):
        _ = ConnackPacket(fixed=header)


@pytest.mark.parametrize("prop", [
    "return_code",
    "session_parent"
])
def test_empty_variable_header(prop):
    packet = ConnackPacket()

    with pytest.raises(ValueError):
        assert getattr(packet, prop) is not None

    with pytest.raises(ValueError):
        assert setattr(packet, prop, "a value")



from dataclasses import dataclass

try:
    from enum import StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  # type: ignore[no-redef]
        pass

import logging
from typing import Any

from aiohttp import ClientResponse, ClientSession, FormData

from amqtt.broker import BrokerContext
from amqtt.contexts import Action
from amqtt.plugins.base import BaseAuthPlugin, BasePlugin, BaseTopicPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


class ResponseMode(StrEnum):
    STATUS = "status"
    JSON = "json"
    TEXT = "text"


class RequestMethod(StrEnum):
    GET = "get"
    POST = "post"
    PUT = "put"


class ParamsMode(StrEnum):
    JSON = "json"
    FORM = "form"


class ACLError(Exception):
    pass


HTTP_2xx_MIN = 200
HTTP_2xx_MAX = 299

HTTP_4xx_MIN = 400
HTTP_4xx_MAX = 499


@dataclass
class HttpConfig:
    """Configuration for the HTTP Auth & ACL Plugin."""

    host: str
    """hostname of the server for the auth & acl check"""
    port: int
    """port of the server for the auth & acl check"""
    request_method: RequestMethod = RequestMethod.GET
    """send the request as a GET, POST or PUT"""
    params_mode: ParamsMode = ParamsMode.JSON  # see docs/plugins/http.md for additional details
    """send the request with `JSON` or `FORM` data. *additional details below*"""
    response_mode: ResponseMode = ResponseMode.JSON  # see docs/plugins/http.md for additional details
    """expected response from the auth/acl server. `STATUS` (code), `JSON`, or `TEXT`. *additional details below*"""
    with_tls: bool = False
    """http or https"""
    user_agent: str = "amqtt"
    """the 'User-Agent' header sent along with the request"""
    superuser_uri: str | None = None
    """URI to verify if the user is a superuser (e.g. '/superuser'), `None` if superuser is not supported"""
    timeout: int = 5
    """duration, in seconds, to wait for the HTTP server to respond"""


class AuthHttpPlugin(BasePlugin[BrokerContext]):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        self.http = ClientSession(headers={"User-Agent": self.config.user_agent})

        match self.config.request_method:
            case RequestMethod.GET:
                self.method = self.http.get
            case RequestMethod.PUT:
                self.method = self.http.put
            case _:
                self.method = self.http.post

    async def on_broker_pre_shutdown(self) -> None:
        await self.http.close()

    @staticmethod
    def _is_2xx(r: ClientResponse) -> bool:
        return HTTP_2xx_MIN <= r.status <= HTTP_2xx_MAX

    @staticmethod
    def _is_4xx(r: ClientResponse) -> bool:
        return HTTP_4xx_MIN <= r.status <= HTTP_4xx_MAX

    def _get_params(self, payload: dict[str, Any]) -> dict[str, Any]:
        match self.config.params_mode:
            case ParamsMode.FORM:
                match self.config.request_method:
                    case RequestMethod.GET:
                        kwargs = {"params": payload}
                    case _:  # POST, PUT
                        d: Any = FormData(payload)
                        kwargs = {"data": d}
            case _:  # JSON
                kwargs = {"json": payload}
        return kwargs

    async def _send_request(self, url: str, payload: dict[str, Any]) -> bool | None:  # pylint: disable=R0911

        kwargs = self._get_params(payload)

        async with self.method(url, **kwargs) as r:
            logger.debug(f"http request returned {r.status}")

            match self.config.response_mode:
                case ResponseMode.TEXT:
                    return self._is_2xx(r) and (await r.text()).lower() == "ok"
                case ResponseMode.STATUS:
                    if self._is_2xx(r):
                        return True
                    if self._is_4xx(r):
                        return False
                    # any other code
                    return None
                case _:
                    if not self._is_2xx(r):
                        return False
                    data: dict[str, Any] = await r.json()
                    data = {k.lower(): v for k, v in data.items()}
                    return data.get("ok", None)

    def get_url(self, uri: str) -> str:
        return f"{'https' if self.config.with_tls else 'http'}://{self.config.host}:{self.config.port}{uri}"


class UserAuthHttpPlugin(AuthHttpPlugin, BaseAuthPlugin):

    async def authenticate(self, *, session: Session) -> bool | None:
        d = {"username": session.username, "password": session.password, "client_id": session.client_id}
        return await self._send_request(self.get_url(self.config.user_uri), d)

    @dataclass
    class Config(HttpConfig):
        """Configuration for the HTTP Auth Plugin."""

        user_uri: str = "/user"
        """URI of the auth check."""


class TopicAuthHttpPlugin(AuthHttpPlugin, BaseTopicPlugin):

    async def topic_filtering(self, *,
                        session: Session | None = None,
                        topic: str | None = None,
                        action: Action | None = None) -> bool | None:
        if not session:
            return None
        acc = 0
        match action:
            case Action.PUBLISH:
                acc = 2
            case Action.SUBSCRIBE:
                acc = 4
            case Action.RECEIVE:
                acc = 1

        d = {"username": session.username, "client_id": session.client_id, "topic": topic, "acc": acc}
        return await self._send_request(self.get_url(self.config.topic_uri), d)

    @dataclass
    class Config(HttpConfig):
        """Configuration for the HTTP Topic Plugin."""

        topic_uri: str = "/acl"
        """URI of the topic check."""



from __future__ import annotations

from importlib import import_module
import logging
from pathlib import Path
import secrets
import string
import sys
import typing
from typing import Any

import yaml

if typing.TYPE_CHECKING:
    from amqtt.session import Session

logger = logging.getLogger(__name__)


def format_client_message(
    session: Session | None = None,
    address: str | None = None,
    port: int | None = None,
) -> str:
    """Format a client message for logging."""
    if session:
        return f"(client id={session.client_id})"
    if address is not None and port is not None:
        return f"(client @={address}:{port})"
    return "(unknown client)"


def gen_client_id() -> str:
    """Generate a random client ID."""
    gen_id = "amqtt/"

    # Use secrets to generate a secure random client ID
    # Defining a valid set of characters for client ID generation
    valid_chars = string.ascii_letters + string.digits
    gen_id += "".join(secrets.choice(valid_chars) for _ in range(16))
    return gen_id


def read_yaml_config(config_file: str | Path) -> dict[str, Any] | None:
    """Read a YAML configuration file."""
    try:
        with Path(str(config_file)).open(encoding="utf-8") as stream:
            yaml_result: dict[str, Any] = yaml.full_load(stream)
            return yaml_result
    except yaml.YAMLError:
        logger.exception(f"Invalid config_file {config_file}")
        return None


def cached_import(module_path: str, class_name: str | None = None) -> Any:
    """Return cached import of a class from a module path (or retrieve, cache and then return)."""
    # Check whether module is loaded and fully initialized.
    if not ((module := sys.modules.get(module_path))
            and (spec := getattr(module, "__spec__", None))
            and getattr(spec, "_initializing", False) is False):
        module = import_module(module_path)
    if class_name:
        return getattr(module, class_name)
    return module


def import_string(dotted_path: str) -> Any:
    """Import a dotted module path.

    Returns:
         attribute/class designated by the last name in the path

    Raises:
         ImportError (if the import failed)

    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        msg = f"{dotted_path} doesn't look like a module path"
        raise ImportError(msg) from err

    try:
        return cached_import(module_path, class_name)
    except AttributeError as err:
        msg = f'Module "{module_path}" does not define a "{class_name}" attribute/class'

        raise ImportError(msg) from err



import asyncio
import logging
import os
import signal
import subprocess
import tempfile
from unittest.mock import patch

import pytest
import yaml

from amqtt.broker import Broker
from amqtt.mqtt.constants import QOS_0

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.DEBUG, format=formatter)
logger = logging.getLogger(__name__)


from amqtt.client import MQTTClient


@pytest.fixture
def broker_config():
    return {
        "listeners": {
            "default": {
                "type": "tcp",
                "bind": "127.0.0.1:1884",  # Use non-standard port for testing
            },
        },
        "sys_interval": 0,
        "auth": {
            "allow-anonymous": True,
            "plugins": ["auth_anonymous"],
        },
        "topic-check": {"enabled": False},
    }


@pytest.fixture
def broker_config_file(broker_config, tmp_path):
    config_path = tmp_path / "broker.yaml"
    with config_path.open("w") as f:
        yaml.dump(broker_config, f)
    return str(config_path)


@pytest.fixture
async def broker(broker_config_file):

    proc = subprocess.Popen(
        ["amqtt", "-c", broker_config_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Give broker time to start
    await asyncio.sleep(4)
    yield proc
    proc.terminate()
    proc.wait()


def test_cli_help_messages():
    """Test that help messages are displayed correctly."""
    env = os.environ.copy()
    env["NO_COLOR"] = '1'

    amqtt_path = "amqtt"
    output = subprocess.check_output([amqtt_path, "--help"], env=env, text=True)
    assert "Usage: amqtt" in output


    amqtt_sub_path = "amqtt_sub"
    output = subprocess.check_output([amqtt_sub_path, "--help"], env=env, text=True)
    assert "Usage: amqtt_sub" in output


    amqtt_pub_path = "amqtt_pub"
    output = subprocess.check_output([amqtt_pub_path, "--help"], env=env, text=True)
    assert "Usage: amqtt_pub" in output


def test_broker_version():
    """Test broker version command."""
    output = subprocess.check_output(["amqtt", "--version"])
    assert output.strip()


@pytest.mark.asyncio
async def test_broker_start_stop(broker_config_file):
    """Test broker start and stop with config file."""
    proc = subprocess.Popen(
        ["amqtt", "-c", broker_config_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    # Give broker time to start
    await asyncio.sleep(1)

    # Verify broker is running by connecting a client
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1:1884")
    await client.disconnect()
    
    # Stop broker
    proc.terminate()
    proc.wait()


@pytest.mark.asyncio
async def test_publish_subscribe(broker):
    """Test pub/sub CLI tools with running broker."""

    # Create a temporary file with test message
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp:
        tmp.write("test message\n")
        tmp.write("another message\n")
        tmp_path = tmp.name

    # Start subscriber in background
    sub_proc = subprocess.Popen(
        [
            "amqtt_sub",
            "--url", "mqtt://127.0.0.1:1884",
            "-t", "test/topic",
            "-n", "2",  # Exit after 2 messages
            "--qos", "1",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    # Give subscriber time to connect
    await asyncio.sleep(0.5)
    #
    # # Publish messages from file
    pub_proc = subprocess.run(
        [
            "amqtt_pub",
            "--url", "mqtt://127.0.0.1:1884",
            "-t", "test/topic",
            "-f", tmp_path,
            "--qos", "1",
        ],
        capture_output=True,
    )
    assert pub_proc.returncode == 0
    #
    # # Wait for subscriber to receive messages
    stdout, stderr = sub_proc.communicate(timeout=5)
    
    # Clean up temp file
    os.unlink(tmp_path)

    # Verify messages were received
    print(stdout.decode("utf-8"))
    assert "test message" in str(stdout)
    assert "another message" in str(stdout)
    assert sub_proc.returncode == 0


@pytest.mark.asyncio
async def test_pub_errors(client_config_file):
    """Test error handling in pub/sub tools."""
    # Test connection to non-existent broker
    cmd =         [
            "amqtt_pub",
            "--url", "mqtt://127.0.0.1:9999",  # Wrong port
            "-t", "test/topic",
            "-m", "test",
            "-c", client_config_file,
        ]

    proc = await asyncio.create_subprocess_shell(
        " ".join(cmd), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await proc.communicate()
    logger.debug(f"Command: {cmd}")
    logger.debug(f"Stdout: {stdout.decode()}")
    logger.debug(f"Stderr: {stderr.decode()}")

    assert proc.returncode != 0, f"publisher error code: {proc.returncode}"
    assert "Connection failed" in str(stderr)


@pytest.mark.asyncio
async def test_sub_errors(client_config_file):
    # Test invalid URL format
    sub_proc = subprocess.run(
        [
            "amqtt_sub",
            "--url", "invalid://url",
            "-t", "test/topic",
            "-c", client_config_file
        ],
        capture_output=True,
    )
    assert sub_proc.returncode != 0, f"subscriber error code: {sub_proc.returncode}"


@pytest.fixture
def client_config():
    return {
        "keep_alive": 10,
        "ping_delay": 1,
        "default_qos": 0,
        "default_retain": False,
        "auto_reconnect": False,
        "will": {
            "topic": "test/will/topic",
            "message": "client ABC has disconnected",
            "qos": 0,
            "retain": False
        },
        "broker": {
            "uri": "mqtt://localhost:1884"
        }
    }


@pytest.fixture
def client_config_file(client_config, tmp_path):
    config_path = tmp_path / "client.yaml"
    with config_path.open("w") as f:
        yaml.dump(client_config, f)
    return str(config_path)


@pytest.mark.asyncio
async def test_pub_client_config(broker, client_config_file):
    await asyncio.sleep(1)
    cmd = [
            "amqtt_pub",
            "-t", "test/topic",
            "-m", "test",
            "-c", client_config_file
        ]
    proc = await asyncio.create_subprocess_shell(
        " ".join(cmd), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await proc.communicate()
    logger.debug(f"Command: {cmd}")
    logger.debug(f"Stdout: {stdout.decode()}")
    logger.debug(f"Stderr: {stderr.decode()}")

    assert proc.returncode == 0, f"publisher error code: {proc.returncode}"



from amqtt.errors import AMQTTError
from amqtt.mqtt.packet import DISCONNECT, MQTTFixedHeader, MQTTPacket


class DisconnectPacket(MQTTPacket[None, None, MQTTFixedHeader]):
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader | None = None) -> None:
        if fixed is None:
            header = MQTTFixedHeader(DISCONNECT, 0x00)
        else:
            if fixed.packet_type is not DISCONNECT:
                msg = f"Invalid fixed packet type {fixed.packet_type} for DisconnectPacket init"
                raise AMQTTError(msg)
            header = fixed
        super().__init__(header)
        self.variable_header = None
        self.payload = None



import asyncio
import unittest
import pytest

from amqtt.adapters import BufferReader
from amqtt.errors import AMQTTError
from amqtt.mqtt.packet import MQTTFixedHeader, CONNECT
from amqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from amqtt.mqtt.publish import PublishPacket, PublishPayload, PublishVariableHeader


class PublishPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream_qos_0(self):
        data = b"\x31\x11\x00\x05topic0123456789"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(PublishPacket.from_stream(stream))
        assert message.variable_header.topic_name == "topic"
        assert message.variable_header.packet_id is None
        assert not message.fixed_header.flags >> 1 & 3
        assert message.fixed_header.flags & 0x01
        assert message.payload.data, b"0123456789"

    def test_from_stream_qos_2(self):
        data = b"\x37\x13\x00\x05topic\x00\x0a0123456789"
        stream = BufferReader(data)
        message = self.loop.run_until_complete(PublishPacket.from_stream(stream))
        assert message.variable_header.topic_name == "topic"
        assert message.variable_header.packet_id == 10
        assert (message.fixed_header.flags >> 1) & 0x03
        assert message.fixed_header.flags & 0x01
        assert message.payload.data, b"0123456789"

    def test_to_stream_no_packet_id(self):
        variable_header = PublishVariableHeader("topic", None)
        payload = PublishPayload(b"0123456789")
        publish = PublishPacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        assert out == b"0\x11\x00\x05topic0123456789"

    def test_to_stream_packet(self):
        variable_header = PublishVariableHeader("topic", 10)
        payload = PublishPayload(b"0123456789")
        publish = PublishPacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        assert out == b"0\x13\x00\x05topic\x00\n0123456789"

    def test_build(self):
        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_0, False)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_0
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_1, False)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_1
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_2, False)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_2
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_0, False)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_0
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_1, False)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_1
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_2, False)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_2
        assert not packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_0, True)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_0
        assert packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_1, True)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_1
        assert packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, False, QOS_2, True)
        assert packet.packet_id == 1
        assert not packet.dup_flag
        assert packet.qos == QOS_2
        assert packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_0, True)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_0
        assert packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_1, True)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_1
        assert packet.retain_flag

        packet = PublishPacket.build("/topic", b"data", 1, True, QOS_2, True)
        assert packet.packet_id == 1
        assert packet.dup_flag
        assert packet.qos == QOS_2
        assert packet.retain_flag


def test_incorrect_fixed_header():
    header = MQTTFixedHeader(CONNECT, 0x00)
    with pytest.raises(AMQTTError):
        _ = PublishPacket(fixed=header)

def test_set_flags():
    packet = PublishPacket()
    packet.set_flags(dup_flag=True, qos=QOS_1, retain_flag=True)


@pytest.mark.parametrize("prop", [
    "packet_id",
    "data",
    "topic_name"
])
def test_empty_variable_header(prop):
    packet = PublishPacket()

    with pytest.raises(ValueError):
        assert getattr(packet, prop) is not None

    with pytest.raises(ValueError):
        assert setattr(packet, prop, "a value")
